from __future__ import annotations

import logging
from typing import Tuple, Optional, Sequence, Iterable, Any, Union, Dict, List

from sqlalchemy import select, func, text, delete, and_, literal
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session, lazyload, joinedload, load_only, noload
from sqlalchemy.orm.attributes import flag_modified

from qcfractal.components.dataset_db_models import BaseDatasetORM, DatasetInternalJobORM, DatasetRecordCountORM
from qcfractal.components.internal_jobs.status import JobProgress
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.services.db_models import ServiceQueueORM
from qcfractal.components.tasks.db_models import TaskQueueORM
from qcfractal.db_socket import SQLAlchemySocket, BaseORM
from qcfractal.db_socket.helpers import get_general, get_query_proj_options, get_count
from qcportal.dataset_models import DatasetModifyMetadata
from qcportal.exceptions import MissingDataError, AlreadyExistsError, UserReportableError
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata, DeleteMetadata, UpdateMetadata
from qcportal.record_models import PriorityEnum, RecordStatusEnum
from qcportal.serialization import encode_to_json
from qcportal.utils import chunk_iterable, now_at_utc


class BaseDatasetSocket:
    """
    Base class for all dataset sockets

    Since all datasets have very similar structure, much of the work can be handled
    in this base class, as long as this class knows the various ORM types.
    """

    # Must be overridden by the derived classes
    dataset_orm = None
    specification_orm = None
    entry_orm = None
    record_item_orm = None
    record_orm = None

    def __init__(
        self,
        root_socket: SQLAlchemySocket,
    ):
        self.root_socket = root_socket

        # Make sure these were set by the derived classes
        assert self.dataset_orm is not None
        assert self.specification_orm is not None
        assert self.entry_orm is not None
        assert self.record_item_orm is not None

        # Use the identity from the ORM object. This keeps everything consistent
        self.dataset_type = self.dataset_orm.__mapper_args__["polymorphic_identity"]

        self._logger = logging.getLogger(__name__)

    def _add_specification(self, session, specification) -> Tuple[InsertMetadata, Optional[int]]:
        raise NotImplementedError("_add_specification must be overridden by the derived class")

    def _create_entries(self, session, dataset_id, new_entries) -> Sequence[BaseORM]:
        raise NotImplementedError("_create_entries must be overridden by the derived class")

    def _submit(
        self,
        session: Session,
        dataset_id: int,
        entry_orm: Iterable[Any],
        specification_orm: Iterable[Any],
        existing_records: Iterable[Tuple[str, str]],
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
    ) -> InsertCountsMetadata:
        raise NotImplementedError("_submit must be overridden by the derived class")

    def get_submit_info(
        self,
        dataset_id: int,
        compute_tag: Optional[str],
        compute_priority: Optional[str],
        creator_user: Optional[Union[int, str]],
        *,
        session: Optional[Session] = None,
    ) -> Tuple[str, PriorityEnum, int]:
        """
        Obtains the tag, priority, and group a new record should be computed with

        This takes into account the fields passed in, as well as the dataset's default
        and priority.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        compute_tag
            Specified tag to use. If None, then the default tag will be used instead
        compute_priority
            Specified priority to use. If None, then the default priority will be used instead
        creator_user
            Name or ID of the user who submitted the dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            The tag, priority, and creator id
        """

        with self.root_socket.optional_session(session, True) as session:
            default_tag, default_priority = self.get_submit_defaults(dataset_id, session=session)

            if compute_tag is None:
                compute_tag = default_tag
            else:
                compute_tag = compute_tag.lower()

            if compute_priority is None:
                compute_priority = default_priority

            user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

        return compute_tag, compute_priority, user_id

    def get_submit_defaults(self, dataset_id: int, *, session: Optional[Session] = None) -> Tuple[str, PriorityEnum]:
        """
        Obtain a dataset's default submission information

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        stmt = select(BaseDatasetORM.default_compute_tag, BaseDatasetORM.default_compute_priority)
        stmt = stmt.where(BaseDatasetORM.id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            r = session.execute(stmt).one_or_none()
            if r is None:
                raise MissingDataError(f"Cannot get default submission info - dataset id {dataset_id} does not exist")
            return tuple(r)

    def get(
        self,
        dataset_id: int,
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Obtain a dataset from the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, None will be returned if the dataset does not exist. Otherwise,
           an exception is raised.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dataset information as a dictionary
        """

        with self.root_socket.optional_session(session) as session:
            return get_general(
                session, self.dataset_orm, self.dataset_orm.id, (dataset_id,), include, exclude, missing_ok
            )[0]

    def status(self, dataset_id: int, *, session: Optional[Session] = None) -> Dict[str, Dict[RecordStatusEnum, int]]:
        """
        Compute the status of a dataset

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dictionary with specifications as the keys, and record status/counts as values.
        """

        stmt = select(self.record_item_orm.specification_name, BaseRecordORM.status, func.count(BaseRecordORM.id))
        stmt = stmt.join(self.record_item_orm, BaseRecordORM.id == self.record_item_orm.record_id)
        stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)
        stmt = stmt.group_by(self.record_item_orm.specification_name, BaseRecordORM.status)

        with self.root_socket.optional_session(session, True) as session:
            stats = session.execute(stmt).all()

        ret: Dict[str, Dict[RecordStatusEnum, int]] = {}
        for s in stats:
            ret.setdefault(s[0], dict())
            ret[s[0]][s[1]] = s[2]
        return ret

    def detailed_status(
        self, dataset_id: int, *, session: Optional[Session] = None
    ) -> List[Tuple[str, str, RecordStatusEnum]]:
        """
        Compute the detailed status of a dataset

        This breaks down the status into entry/specification pairs.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            List of tuple (entry name, specification name, status)
        """

        stmt = select(self.record_item_orm.entry_name, self.record_item_orm.specification_name, BaseRecordORM.status)
        stmt = stmt.join(self.record_item_orm, BaseRecordORM.id == self.record_item_orm.record_id)
        stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            stats = session.execute(stmt).all()
            return [tuple(x) for x in stats]

    def status_by_compute_tag(
        self, dataset_id: int, *, session: Optional[Session] = None
    ) -> List[Tuple[str, RecordStatusEnum, int]]:
        """
        Compute the status of the dataset grouped by tag

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            List of tuple (tag, status, count)
        """

        stmt1 = select(TaskQueueORM.compute_tag, BaseRecordORM.status, func.count(BaseRecordORM.id))
        stmt1 = stmt1.join(BaseRecordORM, TaskQueueORM.record_id == BaseRecordORM.id)
        stmt1 = stmt1.join(self.record_item_orm, BaseRecordORM.id == self.record_item_orm.record_id)
        stmt1 = stmt1.where(self.record_item_orm.dataset_id == dataset_id)
        stmt1 = stmt1.group_by(TaskQueueORM.compute_tag, BaseRecordORM.status)

        stmt2 = select(ServiceQueueORM.compute_tag, BaseRecordORM.status, func.count(BaseRecordORM.id))
        stmt2 = stmt2.join(BaseRecordORM, ServiceQueueORM.record_id == BaseRecordORM.id)
        stmt2 = stmt2.join(self.record_item_orm, BaseRecordORM.id == self.record_item_orm.record_id)
        stmt2 = stmt2.where(self.record_item_orm.dataset_id == dataset_id)
        stmt2 = stmt2.group_by(ServiceQueueORM.compute_tag, BaseRecordORM.status)

        with self.root_socket.optional_session(session, True) as session:
            task_stats = session.execute(stmt1).all()
            service_stats = session.execute(stmt2).all()

            ret: List[Tuple[str, RecordStatusEnum, int]] = []
            for s in task_stats:
                ret.append((s[0], s[1], s[2]))
            for s in service_stats:
                ret.append((s[0], s[1], s[2]))
            return ret

    def get_record_count(self, dataset_id: int, *, session: Optional[Session] = None) -> int:
        """
        Retrieve the number of records stored by the dataset

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Number of records that this dataset contains
        """

        stmt = select(func.count()).select_from(self.record_item_orm)
        stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            count = session.execute(stmt).scalar_one_or_none()

            if count is None:
                raise MissingDataError(f"Could not find dataset with type={self.dataset_type} and id={dataset_id}")

            return count

    def get_computed_properties(self, dataset_id: int, *, session: Optional[Session] = None) -> Dict[str, List[str]]:
        """
        Retrieve the typical properties computed by each specification

        These should be the properties found in the properties dictionary.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Dictionary with keys representing the specification name, and a list of
            properties typically computed by that specification (as a list of strings).
        """

        # Easier to do raw SQL I think
        stmt = text(f"""
             WITH cte AS (
                 SELECT DISTINCT ON (dr.specification_name) dr.specification_name, br.id
                 FROM base_record br
                 INNER JOIN {self.record_item_orm.__tablename__} dr ON dr.record_id = br.id
                 WHERE dr.dataset_id = :dataset_id and br.status = 'complete'
             )
             SELECT cte.specification_name, ARRAY(SELECT jsonb_object_keys(br.properties))
             FROM cte INNER JOIN base_record br ON br.id = cte.id;
        """)

        stmt = stmt.bindparams(dataset_id=dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            rows = session.execute(stmt).all()
            return {r[0]: r[1] for r in rows}

    def add(
        self,
        name: str,
        description: str,
        tagline: str,
        tags: List[str],
        provenance: Dict[str, Any],
        default_compute_tag: str,
        default_compute_priority: PriorityEnum,
        extras: Dict[str, Any],
        creator_user: Optional[Union[int, str]],
        existing_ok: bool,
        *,
        session: Optional[Session] = None,
    ) -> int:
        """
        Create a new dataset in the database

        If a dataset already exists with the same name and type, an exception is raised if existing_ok is False

        Returns
        -------
        :
            ID of the new dataset
        """

        ds_orm = self.dataset_orm(
            dataset_type=self.dataset_type,
            name=name,
            tagline=tagline,
            description=description,
            tags=tags,
            provenance=provenance,
            default_compute_tag=default_compute_tag.lower(),
            default_compute_priority=default_compute_priority,
            extras=extras,
        )

        with self.root_socket.optional_session(session) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user)

            stmt = select(self.dataset_orm.id)
            stmt = stmt.where(self.dataset_orm.lname == name.lower())
            stmt = stmt.where(self.dataset_orm.dataset_type == self.dataset_type)
            existing_id = session.execute(stmt).scalar_one_or_none()

            if existing_id is not None:
                if existing_ok:
                    return existing_id
                else:
                    raise AlreadyExistsError(
                        f"Dataset with type='{self.dataset_type}' and name='{name}' already exists"
                    )

            ds_orm.creator_user_id = creator_user_id

            session.add(ds_orm)
            session.commit()
            return ds_orm.id

    def update_record_count(self, dataset_id: int, change: Optional[int], *, session: Optional[Session] = None):
        """
        Updates the number of records that this dataset has

        This information is used when listing datasets, since dynamically determining
        the number of records is too expensive for large qcfractal instances.

        If change is None, the total number of records will be calculated based on what is in the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        change
            Change (+/-) to apply to the existing number of records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            stmt = select(DatasetRecordCountORM).where(DatasetRecordCountORM.dataset_id == dataset_id).with_for_update()
            rcount_orm = session.execute(stmt).scalar_one_or_none()

            if rcount_orm is None:
                rcount_orm = DatasetRecordCountORM(dataset_id=dataset_id, record_count=0)

            if change is not None:
                rcount_orm.record_count += change
            else:
                stmt = (
                    select(func.count(self.record_item_orm.dataset_id))
                    .where(self.record_item_orm.dataset_id == dataset_id)
                    .group_by(self.record_item_orm.dataset_id)
                )
                rcount_orm.record_count = session.execute(stmt).scalar_one()

            session.add(rcount_orm)

    def update_metadata(
        self, dataset_id: int, new_metadata: DatasetModifyMetadata, *, session: Optional[Session] = None
    ):
        """
        Updates the metadata of the dataset

        This will overwrite the existing metadata. An exception is raised on any error

        Parameters
        ----------
        dataset_id
            ID of a dataset
        new_metadata
            New metadata to store
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            stmt = select(self.dataset_orm).where(self.dataset_orm.id == dataset_id)
            stmt = stmt.options(lazyload("*")).with_for_update()
            ds = session.execute(stmt).scalar_one_or_none()

            if ds is None:
                raise MissingDataError(f"Could not find dataset with type={self.dataset_type} and id={dataset_id}")

            if ds.name != new_metadata.name:
                # If only change in case, no need to check if it already exists
                if ds.name.lower() != new_metadata.name.lower():
                    stmt2 = select(self.dataset_orm.id)
                    stmt2 = stmt2.where(self.dataset_orm.dataset_type == self.dataset_type)
                    stmt2 = stmt2.where(self.dataset_orm.lname == new_metadata.name.lower())
                    existing = session.execute(stmt2).scalar_one_or_none()

                    if existing:
                        raise AlreadyExistsError(
                            f"{self.dataset_type} dataset named '{new_metadata.name}' already exists"
                        )

                ds.name = new_metadata.name

            ds.description = new_metadata.description
            ds.tagline = new_metadata.tagline
            ds.tags = new_metadata.tags
            ds.provenance = new_metadata.provenance
            ds.extras = new_metadata.extras

            ds.default_compute_tag = new_metadata.default_compute_tag
            ds.default_compute_priority = new_metadata.default_compute_priority

    def add_specifications(
        self,
        dataset_id: int,
        new_specifications: Sequence[Any],  # we don't know the type here
        *,
        session: Optional[Session] = None,
    ) -> InsertMetadata:
        """
        Adds specifications to a dataset in the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        new_specifications
            Specifications to add
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion
        """

        existing_idx: List[int] = []
        inserted_idx: List[int] = []
        errors: List[Tuple[int, str]] = []

        with self.root_socket.optional_session(session) as session:
            stmt = select(self.specification_orm.name)
            stmt = stmt.where(self.specification_orm.dataset_id == dataset_id)

            existing_specs = session.execute(stmt).scalars().all()

            for idx, ds_spec in enumerate(new_specifications):
                if ds_spec.name in existing_specs:
                    existing_idx.append(idx)
                    continue

                # call the derived class function for adding a specification
                meta, spec_id = self._add_specification(session, ds_spec.specification)

                if not meta.success:
                    err_str = f"Unable to add {self.dataset_type} specification: " + meta.error_string
                    errors.append((idx, err_str))
                    continue

                ds_spec_orm = self.specification_orm(
                    dataset_id=dataset_id, name=ds_spec.name, description=ds_spec.description, specification_id=spec_id
                )

                session.add(ds_spec_orm)
                inserted_idx.append(idx)

        return InsertMetadata(inserted_idx=inserted_idx, existing_idx=existing_idx)

    def fetch_specification_names(
        self,
        dataset_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[str]:
        """
        Obtain all specification names for a dataset

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            All entry names as a list
        """
        stmt = select(self.specification_orm.name)
        stmt = stmt.where(self.specification_orm.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            ret = session.execute(stmt).scalars().all()
            return list(ret)

    def fetch_specifications(
        self,
        dataset_id: int,
        specification_names: Optional[Sequence[str]] = None,
        include: Optional[Iterable[str]] = None,
        exclude: Optional[Iterable[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        """
        Fetch specifications for a dataset from the database

        It's expected there aren't too many specifications, so this always fetches them all.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        specification_names
            Names of the specifications to fetch. If None, fetch all
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Specifications as a dictionary, with the key being the name and the value being
            the specification (as a dictionary)
        """

        stmt = select(self.specification_orm)
        stmt = stmt.where(self.specification_orm.dataset_id == dataset_id)
        stmt = stmt.options(joinedload(self.specification_orm.specification))

        if specification_names is not None:
            stmt = stmt.where(self.specification_orm.name.in_(specification_names))

        if include or exclude:
            query_opts = get_query_proj_options(self.specification_orm, include, exclude)
            stmt = stmt.options(*query_opts)

        with self.root_socket.optional_session(session, True) as session:
            specifications = session.execute(stmt).scalars().all()

        if specification_names is not None and missing_ok is False:
            found_specifications = {x.name for x in specifications}
            missing_specifications = set(specification_names) - found_specifications
            if missing_specifications:
                s = "\n".join(missing_specifications)
                raise MissingDataError(f"Missing {len(missing_specifications)} specifications: {s}")

        return {x.name: x.model_dict() for x in specifications}

    def delete_specifications(
        self,
        dataset_id: int,
        specification_names: Iterable[str],
        delete_records: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> DeleteMetadata:
        """
        Deletes specifications from a dataset

        Specifications which do not exist for the dataset will be silently ignored

        Parameters
        ----------
        dataset_id
            ID of a dataset
        specification_names
            Names of the specifications to delete
        delete_records
            If True, also (hard) delete all records associated with this specification
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about the deletion. If delete_records is True, then n_children_deleted will
            contain the number of records deleted.
        """

        with self.root_socket.optional_session(session) as session:
            if delete_records:
                # Store all record ids for later deletion
                stmt = select(self.record_item_orm.record_id)
                stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)
                stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))
                record_ids = session.execute(stmt).scalars().all()

            # Deleting the specification will cascade to the dataset->record association table
            stmt = delete(self.specification_orm)
            stmt = stmt.where(self.specification_orm.dataset_id == dataset_id)
            stmt = stmt.where(self.specification_orm.name.in_(specification_names))
            stmt = stmt.returning(self.specification_orm.name)
            deleted_entries = session.execute(stmt).scalars().all()
            session.flush()

            n_children_deleted = 0
            if delete_records:
                rec_meta = self.root_socket.records.delete(
                    record_ids, soft_delete=False, delete_children=True, session=session
                )
                n_children_deleted = rec_meta.n_deleted

            deleted_idx = [idx for idx, name in enumerate(specification_names) if name in deleted_entries]
            errors = [
                (idx, "specification does not exist")
                for idx, name in enumerate(specification_names)
                if name not in deleted_entries
            ]
            return DeleteMetadata(deleted_idx=deleted_idx, errors=errors, n_children_deleted=n_children_deleted)

    def rename_specifications(
        self, dataset_id: int, specification_name_map: Dict[str, str], *, session: Optional[Session] = None
    ):
        """
        Renames specifications

        The specification_name_map maps the old name to the new name (ie, `specification_name_map[old_name] = new_name`).
        If any of the new names exist, an exception is raised and no renaming takes place.
        If a specification does not exist under the old name, then that renaming is silently ignored.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        specification_name_map
            Mapping of old name to new name
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        specification_name_map = {k: v for k, v in specification_name_map.items() if k != v}
        if not specification_name_map:
            return

        stmt = select(self.specification_orm)
        stmt = stmt.where(self.specification_orm.dataset_id == dataset_id)
        stmt = stmt.where(self.specification_orm.name.in_(specification_name_map.keys()))
        stmt = stmt.options(load_only(self.specification_orm.name))

        # See if any of the new names already exist
        exist_stmt = select(self.specification_orm.name)
        exist_stmt = exist_stmt.where(self.specification_orm.dataset_id == dataset_id)
        exist_stmt = exist_stmt.where(self.specification_orm.name.in_(specification_name_map.values()))

        with self.root_socket.optional_session(session) as session:
            existing = session.execute(exist_stmt).scalars().all()
            if existing:
                raise AlreadyExistsError(
                    f"Cannot rename specification to {existing[0]} - specification with that name already exists"
                )

            specs = session.execute(stmt).scalars().all()

            for spec in specs:
                spec.name = specification_name_map[spec.name]

    def add_entries(
        self, dataset_id: int, new_entries: Sequence[Any], *, session: Optional[Session] = None
    ) -> InsertMetadata:
        """
        Adds entries to a dataset in the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        new_entries
            Entries to add
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion
        """

        with self.root_socket.optional_session(session) as session:
            # Create orm for all entries (in derived class)
            entry_orm = self._create_entries(session, dataset_id, new_entries)

            inserted_idx: List[int] = []
            existing_idx: List[int] = []

            # Go in batches of 200 to avoid huge queries
            idx = 0
            for entry_orm_batch in chunk_iterable(entry_orm, 200):
                # Get all existing entries first
                stmt = select(self.entry_orm.name)
                stmt = stmt.where(
                    and_(
                        self.entry_orm.dataset_id == dataset_id,
                        self.entry_orm.name.in_([x.name for x in entry_orm_batch]),
                    )
                )

                existing_entries = set(session.execute(stmt).scalars().all())
                entries_to_add = []

                for entry in entry_orm_batch:
                    # Only add if the entry does not exist
                    if entry.name in existing_entries:
                        existing_idx.append(idx)
                    else:
                        entries_to_add.append(entry)
                        inserted_idx.append(idx)

                    idx += 1

                session.add_all(entries_to_add)

        return InsertMetadata(inserted_idx=inserted_idx, existing_idx=existing_idx)

    def background_add_entries(
        self, dataset_id: int, new_entries: Sequence[Any], *, session: Optional[Session] = None
    ) -> int:
        """
        Adds entries to a dataset in the database as an internal job

        This creates an internal job for the addition and returns the ID

        See :meth:`add_entries` for details for the rest of the details on functionality and parameters.

        Returns
        -------
        :
            ID of the created internal job
        """

        with self.root_socket.optional_session(session) as session:
            job_id = self.root_socket.internal_jobs.add(
                f"dataset_add_entries_{dataset_id}",
                now_at_utc(),
                f"datasets.add_entry_dicts",
                {
                    "dataset_id": dataset_id,
                    "entry_dicts": encode_to_json(new_entries),
                },
                user_id=None,
                unique_name=False,
                serial_group=f"ds_add_entries_{dataset_id}",  # only run one addition for this dataset at a time
                session=session,
            )

            stmt = (
                insert(DatasetInternalJobORM)
                .values(dataset_id=dataset_id, internal_job_id=job_id)
                .on_conflict_do_nothing()
            )
            session.execute(stmt)
            return job_id

    def fetch_entry_names(
        self,
        dataset_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[str]:
        """
        Obtain all entry names for a dataset

        Parameters
        ----------
        dataset_id
            ID of a dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            All entry names as a list
        """
        stmt = select(self.entry_orm.name)
        stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)

        with self.root_socket.optional_session(session, True) as session:
            ret = session.execute(stmt).scalars().all()
            return list(ret)

    def fetch_entries(
        self,
        dataset_id: int,
        entry_names: Optional[Sequence[str]] = None,
        include: Optional[Iterable[str]] = None,
        exclude: Optional[Iterable[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        """
        Obtain full entries for a dataset from the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        entry_names
            Names of the entries to fetch. If None, fetch all
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, None will be returned if the dataset does not exist. Otherwise,
           an exception is raised.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Entries as a dictionary, with the key being the name and the value being
            the entry (as a dictionary)
        """

        stmt = select(self.entry_orm)
        stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)

        if entry_names is not None:
            stmt = stmt.where(self.entry_orm.name.in_(entry_names))

        if include or exclude:
            query_opts = get_query_proj_options(self.entry_orm, include, exclude)
            stmt = stmt.options(*query_opts)

        with self.root_socket.optional_session(session, True) as session:
            entries = session.execute(stmt).scalars().all()

        if entry_names is not None and missing_ok is False:
            found_entries = {x.name for x in entries}
            missing_entries = set(entry_names) - found_entries
            if missing_entries:
                s = "\n".join(missing_entries)
                raise MissingDataError(f"Missing {len(missing_entries)} entries: {s}")

        return {x.name: x.model_dict() for x in entries}

    def delete_entries(
        self,
        dataset_id: int,
        entry_names: Sequence[str],
        delete_records: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> DeleteMetadata:
        """
        Deletes entries from a dataset

        Entries which do not exist for the dataset will be silently ignored

        Parameters
        ----------
        dataset_id
            ID of a dataset
        entry_names
            Names of the entries to delete
        delete_records
            If True, also (hard) delete all records associated with this specification
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about the deletion. If delete_records is True, then n_children_deleted will
            contain the number of records deleted.
        """

        with self.root_socket.optional_session(session) as session:
            if delete_records:
                # Store all record ids for later deletion
                stmt = select(self.record_item_orm.record_id)
                stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)
                stmt = stmt.where(self.record_item_orm.entry_name.in_(entry_names))
                record_ids = session.execute(stmt).scalars().all()

            # Delete the entries will cascade to the dataset->record association table
            stmt = delete(self.entry_orm)
            stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)
            stmt = stmt.where(self.entry_orm.name.in_(entry_names))
            stmt = stmt.returning(self.entry_orm.name)
            deleted_entries = session.execute(stmt).scalars().all()
            session.flush()

            n_children_deleted = 0
            if delete_records:
                rec_meta = self.root_socket.records.delete(
                    record_ids, soft_delete=False, delete_children=True, session=session
                )
                n_children_deleted = rec_meta.n_deleted

            deleted_idx = [idx for idx, name in enumerate(entry_names) if name in deleted_entries]
            errors = [
                (idx, "entry does not exist") for idx, name in enumerate(entry_names) if name not in deleted_entries
            ]
            return DeleteMetadata(deleted_idx=deleted_idx, errors=errors, n_children_deleted=n_children_deleted)

    def rename_entries(self, dataset_id: int, entry_name_map: Dict[str, str], *, session: Optional[Session] = None):
        """
        Renames entries for a dataset

        The entry_name_map maps the old name to the new name (ie, `entry_name_map[old_name] = new_name`).
        If any of the new names exist, an exception is raised and no renaming takes place.
        If an entry does not exist under the old name, then that renaming is silently ignored.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        entry_name_map
            Mapping of old name to new name
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        # strip out any renames that don't rename
        entry_name_map = {k: v for k, v in entry_name_map.items() if k != v}
        if not entry_name_map:
            return

        stmt = select(self.entry_orm)
        stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)
        stmt = stmt.where(self.entry_orm.name.in_(entry_name_map.keys()))
        stmt = stmt.options(load_only(self.entry_orm.name))

        # See if any of the new names already exist
        exist_stmt = select(self.entry_orm.name)
        exist_stmt = exist_stmt.where(self.entry_orm.dataset_id == dataset_id)
        exist_stmt = exist_stmt.where(self.entry_orm.name.in_(entry_name_map.values()))

        with self.root_socket.optional_session(session) as session:
            existing = session.execute(exist_stmt).scalars().all()
            if existing:
                raise AlreadyExistsError(f"Cannot rename entry to {existing[0]} - entry with that name already exists")

            # Now do the renaming
            entries = session.execute(stmt).scalars().all()

            for entry in entries:
                entry.name = entry_name_map[entry.name]

    def modify_entries(
        self,
        dataset_id: int,
        attribute_map: Optional[Dict[str, Dict[str, Any]]] = None,
        comment_map: Optional[Dict[str, str]] = None,
        overwrite_attributes: bool = False,
        *,
        session: Optional[Session] = None,
    ):
        """
        Modify the attributes of the entries in a dataset.

        If overwrite_attributes is True, replaces existing attribute entry with the value in attribute_map.
        If overwrite_attributes is False, updates existing fields within attributes and adds non-existing fields.
        The attribute_map maps the name of the entry to the new attribute data.
        The comment_map maps the name of an entry to the comment.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        attribute_map
            Mapping of entry names to attributes.
        comment_map
            Mapping of entry names to comments
        overwrite_attributes
            Boolean to indicate if existing entries should be overwritten.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """
        stmt = select(self.entry_orm)
        stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)

        stmt = stmt.where(
            self.entry_orm.name.in_(
                (attribute_map.keys() if (attribute_map is not None) else set())
                | (comment_map.keys() if (comment_map is not None) else set())
            )
        )
        stmt = stmt.options(load_only(self.entry_orm.name, self.entry_orm.attributes, self.entry_orm.comment))
        stmt = stmt.options(lazyload("*"))
        stmt = stmt.with_for_update(skip_locked=False)

        attribute_keys = attribute_map.keys() if (attribute_map is not None) else list()
        comment_keys = comment_map.keys() if (comment_map is not None) else list()

        with self.root_socket.optional_session(session) as session:
            entries = session.execute(stmt).scalars().all()

            for entry in entries:
                if overwrite_attributes:
                    if entry.name in attribute_keys:
                        entry.attributes = attribute_map[entry.name]
                else:
                    if entry.name in attribute_keys:
                        entry.attributes.update(attribute_map[entry.name])
                        flag_modified(entry, "attributes")

                if entry.name in comment_keys:
                    entry.comment = comment_map[entry.name]

    def fetch_records(
        self,
        dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
        specification_names: Optional[Iterable[str]] = None,
        status: Optional[Iterable[RecordStatusEnum]] = None,
        *,
        session: Optional[Session] = None,
    ) -> List[Tuple[str, str, int]]:
        """
        Obtain record ids for a dataset from the database

        The returned list is in an indeterminant order

        Parameters
        ----------
        dataset_id
            ID of a dataset
        entry_names
            Fetch records belonging to these entries. If None, fetch records belonging to any entry.
        specification_names
            Fetch records belonging to these specifications. If None, fetch records belonging to any specification.
        status
            Fetch records whose status is in the given list (or other iterable) of statuses
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of record information in the form (entry_name, specification_name, record_id)
        """

        stmt = select(self.record_item_orm)
        stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)

        if entry_names is not None:
            stmt = stmt.where(self.record_item_orm.entry_name.in_(entry_names))
        if specification_names is not None:
            stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))
        if status:
            stmt = stmt.join(self.record_item_orm.record)
            stmt = stmt.where(self.record_orm.status.in_(status))

        with self.root_socket.optional_session(session, True) as session:
            record_items = session.execute(stmt).scalars().all()
            return [(x.entry_name, x.specification_name, x.record_id) for x in record_items]

    def remove_records(
        self,
        dataset_id: int,
        entry_names: Sequence[str],
        specification_names: Sequence[str],
        delete_records: bool,
        *,
        session: Optional[Session] = None,
    ):
        """
        Removes a record from this dataset, optionally deleting the record

        This does not delete entries or specifications.

        Parameters
        ----------
        dataset_id
            ID of a dataset
        entry_names
            Delete records belonging to these entries
        specification_names
            Delete records belonging to these specifications
        delete_records
            If True, actually delete the records (rather than simply un-associating them from this dataset)
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            if delete_records:
                # Store all record ids for later deletion
                stmt = select(self.record_item_orm.record_id)
                stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)
                stmt = stmt.where(self.record_item_orm.entry_name.in_(entry_names))
                stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))
                record_ids = session.execute(stmt).scalars().all()

            stmt = delete(self.record_item_orm)
            stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)
            stmt = stmt.where(self.record_item_orm.entry_name.in_(entry_names))
            stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))
            r = session.execute(stmt)

            if delete_records:
                self.root_socket.records.delete(record_ids, soft_delete=False, delete_children=True, session=session)

            self.update_record_count(dataset_id, -r.rowcount, session=session)

    def delete_dataset(
        self,
        dataset_id: int,
        delete_records: bool,
        *,
        session: Optional[Session] = None,
    ):
        """
        Deletes an entire dataset from the database

        Parameters
        ----------
        dataset_id
            ID of a dataset
        delete_records
            If true, delete all the individual records as well
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            if delete_records:
                # Store all record ids for later deletion
                stmt = select(self.record_item_orm.record_id)
                stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)
                record_ids = session.execute(stmt).scalars().all()

            stmt = delete(BaseDatasetORM)
            stmt = stmt.where(BaseDatasetORM.id == dataset_id)
            session.execute(stmt)

            if delete_records:
                self.root_socket.records.delete(record_ids, soft_delete=False, delete_children=True, session=session)

    def submit(
        self,
        dataset_id: int,
        entry_names: Optional[Iterable[str]],
        specification_names: Optional[Iterable[str]],
        compute_tag: Optional[str],
        compute_priority: Optional[PriorityEnum],
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        job_progress: Optional[JobProgress] = None,
        session: Optional[Session] = None,
    ) -> InsertCountsMetadata:
        """
        Submit computations for this dataset

        If any specification or entry name does not exist, an exception is raised and
        no submission takes place

        Parameters
        ----------
        dataset_id
            ID of a dataset
        entry_names
            Submit records belonging to these entries. If None, submit for all entries
        specification_names
            Submit records belonging to these specifications. If None, submit for all specifications
        compute_tag
            Computational tag for new records. If None, use the dataset's default. Existing records
            will not be modified.
        compute_priority
            Priority for new records. If None, use the dataset's default. Existing records
            will not be modified.
        creator_user
            Name or ID of the user who submitted the dataset
        find_existing
            If True, search for existing records and return those. If False, always add new records
        job_progress
            Object used to track progress if this function is being run in a background job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Counts of how many records were inserted or already existing. This only applies to records - existing
            records already part of this dataset (ie, a given entry/specification pair already has a record)
            is not counted as existing in the return value.
        """

        n_inserted = 0
        n_existing = 0

        with self.root_socket.optional_session(session) as session:
            compute_tag, compute_priority, creator_user_id = self.get_submit_info(
                dataset_id, compute_tag, compute_priority, creator_user, session=session
            )

            ################################
            # Get specification details
            ################################
            stmt = select(self.specification_orm)

            # We want the actual full specification as well
            stmt = stmt.join(self.specification_orm.specification)
            stmt = stmt.where(self.specification_orm.dataset_id == dataset_id)
            if specification_names is not None:
                stmt = stmt.where(self.specification_orm.name.in_(specification_names))

            ds_specs = session.execute(stmt).scalars().all()

            # Check to make sure we found all the specifications
            if specification_names is not None:
                found_specs = {x.name for x in ds_specs}
                missing_specs = set(specification_names) - found_specs
                if missing_specs:
                    raise MissingDataError(f"Could not find all specifications. Missing: {missing_specs}")

            ################################
            # Get entry details
            ################################
            if entry_names is None:
                # Do all entries in batches using server-side cursors
                stmt = select(self.entry_orm)
                stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)

                # for progress tracking
                if job_progress is not None:
                    total_records = len(ds_specs) * get_count(session, stmt)
                    records_done = 0

                r = session.execute(stmt).scalars()

                while entries_batch := r.fetchmany(500):
                    entries_batch_names = [e.name for e in entries_batch]

                    # Find which records/record_items already exist
                    stmt = select(self.record_item_orm.entry_name, self.record_item_orm.specification_name)
                    stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)

                    stmt = stmt.where(self.record_item_orm.entry_name.in_(entries_batch_names))
                    if specification_names is not None:
                        stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))

                    existing_records = session.execute(stmt).all()

                    batch_meta = self._submit(
                        session,
                        dataset_id,
                        entries_batch,
                        ds_specs,
                        existing_records,
                        compute_tag,
                        compute_priority,
                        creator_user_id,
                        find_existing,
                    )

                    n_inserted += batch_meta.n_inserted
                    n_existing += batch_meta.n_existing

                    if job_progress is not None:
                        job_progress.raise_if_cancelled()
                        records_done += len(entries_batch)
                        job_progress.update_progress(100 * (records_done * len(ds_specs)) / total_records)

            else:  # entry names were given

                # for progress tracking
                if job_progress is not None:
                    total_records = len(ds_specs) * len(entry_names)
                    records_done = 0

                # For checking for missing entries
                found_entries = []

                # Do entries in batches via the given entry names (in batches)
                for entries_names_batch in chunk_iterable(entry_names, 500):
                    stmt = select(self.entry_orm)
                    stmt = stmt.where(self.entry_orm.dataset_id == dataset_id)
                    stmt = stmt.where(self.entry_orm.name.in_(entries_names_batch))

                    entries_batch = session.execute(stmt).scalars().all()

                    entries_batch_names = [e.name for e in entries_batch]
                    found_entries.extend(entries_batch_names)

                    # Find which records/record_items already exist
                    stmt = select(self.record_item_orm.entry_name, self.record_item_orm.specification_name)
                    stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)

                    stmt = stmt.where(self.record_item_orm.entry_name.in_(entries_batch_names))
                    if specification_names is not None:
                        stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))

                    existing_records = session.execute(stmt).all()

                    batch_meta = self._submit(
                        session,
                        dataset_id,
                        entries_batch,
                        ds_specs,
                        existing_records,
                        compute_tag,
                        compute_priority,
                        creator_user_id,
                        find_existing,
                    )

                    n_inserted += batch_meta.n_inserted
                    n_existing += batch_meta.n_existing

                    if job_progress is not None:
                        job_progress.raise_if_cancelled()
                        records_done += len(entries_names_batch)
                        job_progress.update_progress(100 * (records_done * len(ds_specs)) / total_records)

                if entry_names is not None:
                    missing_entries = set(entry_names) - set(found_entries)
                    if missing_entries:
                        raise MissingDataError(f"Could not find all entries. Missing: {missing_entries}")

            # Update the record count
            # n_inserted and n_existing represent records inserted/existing in the database, NOT this dataset
            # What was already existing in the dataset is not in these values. So the total change
            # of what was linked to this dataset is n_inserted + n_existing
            self.update_record_count(dataset_id, n_inserted + n_existing, session=session)

        return InsertCountsMetadata(n_inserted=n_inserted, n_existing=n_existing)

    def background_submit(
        self,
        dataset_id: int,
        entry_names: Optional[Iterable[str]],
        specification_names: Optional[Iterable[str]],
        compute_tag: Optional[str],
        compute_priority: Optional[PriorityEnum],
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> int:
        """
        Submit computations for this dataset as an internal job

        This creates an internal job for the submission and returns the ID

        See :meth:`submit` for details for the rest of the details on functionality and parameters.

        Returns
        -------
        :
            ID of the created internal job
        """

        with self.root_socket.optional_session(session) as session:
            job_id = self.root_socket.internal_jobs.add(
                f"dataset_submit_{dataset_id}",
                now_at_utc(),
                f"datasets.submit",
                {
                    "dataset_id": dataset_id,
                    "entry_names": entry_names,
                    "specification_names": specification_names,
                    "compute_tag": compute_tag,
                    "compute_priority": compute_priority,
                    "creator_user": creator_user,
                    "find_existing": find_existing,
                },
                user_id=None,
                unique_name=False,
                serial_group=f"ds_submit_{dataset_id}",  # only run one submission for this dataset at a time
                session=session,
            )

            stmt = (
                insert(DatasetInternalJobORM)
                .values(dataset_id=dataset_id, internal_job_id=job_id)
                .on_conflict_do_nothing()
            )
            session.execute(stmt)
            return job_id

    #######################
    # Record modification
    #######################
    def _lookup_record_ids(
        self,
        session: Session,
        dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
        specification_names: Optional[Iterable[str]] = None,
        status: Optional[Iterable[RecordStatusEnum]] = None,
        for_update: bool = False,
    ) -> List[int]:
        """
        Lookup the record IDs for a dataset

        The returned IDs are in an indeterminant order

        Parameters
        ----------
        session
            An existing SQLAlchemy session to use
        dataset_id
            ID of a dataset
        entry_names
            Find records belonging to these entries. If None, look up all entries
        specification_names
            Find records belonging to these specifications. If None, look up all specifications
        status
            Look up records whose status is in the given list (or other iterable) of statuses
        for_update
            If True, select the records with row-level locking (postgres FOR UPDATE)

        Returns
        -------
        :
            List of record IDs
        """
        stmt = select(self.record_item_orm.record_id)
        stmt = stmt.where(self.record_item_orm.dataset_id == dataset_id)

        if entry_names is not None:
            stmt = stmt.where(self.record_item_orm.entry_name.in_(entry_names))
        if specification_names is not None:
            stmt = stmt.where(self.record_item_orm.specification_name.in_(specification_names))
        if status is not None:
            stmt = stmt.join(BaseRecordORM, BaseRecordORM.id == self.record_item_orm.record_id)
            stmt = stmt.where(BaseRecordORM.status.in_(status))

        # Locks the record items, not the actual record
        if for_update:
            stmt = stmt.with_for_update()

        r = session.execute(stmt).scalars().all()
        return r

    def modify_records(
        self,
        dataset_id: int,
        username: Optional[str],
        entry_names: Optional[Iterable[str]] = None,
        specification_names: Optional[List[str]] = None,
        status: Optional[RecordStatusEnum] = None,
        compute_priority: Optional[PriorityEnum] = None,
        compute_tag: Optional[str] = None,
        comment: Optional[str] = None,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Modify records belonging to a dataset

        Parameters
        ----------
        dataset_id
            ID of a dataset
        username
            Username of the user modifying these records
        entry_names
            Modify records belonging to these entries. If None, modify records belonging to any entry.
        specification_names
            Modify records belonging to these specifications. If None, modify records belonging to any specification.
        username
            Username of the user modifying the records
        status
            New status for the records. Only certain status transitions will be allowed.
        compute_priority
            New priority for these records
        compute_tag
            New tag for these records
        comment
            Adds a new comment to these records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about the modification/update.
        """

        with self.root_socket.optional_session(session) as session:
            record_ids = self._lookup_record_ids(
                session,
                dataset_id,
                entry_names,
                specification_names,
                for_update=True,
            )

            return self.root_socket.records.modify_generic(
                record_ids,
                username,
                status=status,
                compute_priority=compute_priority,
                compute_tag=compute_tag,
                comment=comment,
                session=session,
            )

    def revert_records(
        self,
        dataset_id: int,
        revert_status: RecordStatusEnum,
        entry_names: Optional[Iterable[str]] = None,
        specification_names: Optional[List[str]] = None,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Reverts the status of dataset records to their previous status

        Parameters
        ----------
        dataset_id
            ID of a dataset
        revert_status
            Revert records of this status. For example, if `revert_status` is `deleted`, then it will be undeleted
        entry_names
            Modify records belonging to these entries. If None, modify records belonging to any entry.
        specification_names
            Modify records belonging to these specifications. If None, modify records belonging to any specification.

        Returns
        -------
        :
            Metadata about the modification/update

        """

        with self.root_socket.optional_session(session) as session:
            record_ids = self._lookup_record_ids(session, dataset_id, entry_names, specification_names, for_update=True)

            return self.root_socket.records.revert_generic(record_ids, revert_status)

    def _copy_entries(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
    ):
        """
        Copy entries from one dataset to another

        If `entry_names` is not provided, all entries will be copied
        """

        raise NotImplementedError("_clone_entries must be overridden by the derived class")

    def copy_entries(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
    ):
        """
        Copy entries from one dataset to another

        If `entry_names` is not provided, all entries will be copied
        """

        try:
            self._copy_entries(session, source_dataset_id, destination_dataset_id, entry_names)
        except IntegrityError:
            raise UserReportableError(
                "Cannot copy entries from dataset - destination already has entries with the same name"
            )

    def copy_specifications(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        specification_names: Optional[Iterable[str]] = None,
    ):
        """
        Copy specifications from one dataset to another

        If `specification_names` is not provided, all specifications will be copied
        """

        # Which specs already exist in the destination dataset
        # We will allow duplicates if name and spec id match, so we filter those out with an outer join
        existing_cte = select(
            self.specification_orm.name,
            self.specification_orm.specification_id,
        )

        existing_cte = existing_cte.where(self.specification_orm.dataset_id == destination_dataset_id).cte()

        # All dataset specifications (so far) have the same structure. So we can do it for all
        # types of datasets here
        select_stmt = select(
            literal(destination_dataset_id),
            self.specification_orm.name,
            self.specification_orm.description,
            self.specification_orm.specification_id,
        )

        # Join against the CTE for the existing specs
        # If a spec exists with the same name and spec id, then the "where is null" will be false and it will be
        #     filtered out from the select
        # If the name exists BUT THE ID IS DIFFERENT, it is not filtered out, which will cause an integrity error
        #     when we insert later
        select_stmt = select_stmt.join(
            existing_cte,
            and_(
                self.specification_orm.name == existing_cte.c.name,
                self.specification_orm.specification_id == existing_cte.c.specification_id,
            ),
            isouter=True,
        )
        select_stmt = select_stmt.where(self.specification_orm.dataset_id == source_dataset_id)
        select_stmt = select_stmt.where(existing_cte.c.name.is_(None))

        if specification_names is not None:
            select_stmt = select_stmt.where(self.specification_orm.name.in_(specification_names))

        stmt = insert(self.specification_orm)
        stmt = stmt.from_select(
            [
                self.specification_orm.dataset_id,
                self.specification_orm.name,
                self.specification_orm.description,
                self.specification_orm.specification_id,
            ],
            select_stmt,
        )

        try:
            session.execute(stmt)
        except IntegrityError:
            raise UserReportableError(
                "Cannot copy specifications from dataset - destination already has specifications with the same name"
            )

    def copy_record_items(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
        specification_names: Optional[Iterable[str]] = None,
    ):
        """
        Copy record items from one dataset to another

        This doesn't clone actual records, just the links to the records for the dataset

        If `entry_names` is not provided, all entries will be copied
        If `specification_names` is not provided, all entries will be copied.
        """

        # All dataset record items (so far) have the same structure. So we can do it for all
        # types of datasets here
        select_stmt = select(
            literal(destination_dataset_id),
            self.record_item_orm.entry_name,
            self.record_item_orm.specification_name,
            self.record_item_orm.record_id,
        )

        select_stmt = select_stmt.where(self.record_item_orm.dataset_id == source_dataset_id)

        if entry_names is not None:
            select_stmt = select_stmt.where(self.record_item_orm.entry_name.in_(entry_names))
        if specification_names is not None:
            select_stmt = select_stmt.where(self.record_item_orm.specification_name.in_(specification_names))

        stmt = insert(self.record_item_orm)
        stmt = stmt.from_select(
            [
                self.record_item_orm.dataset_id,
                self.record_item_orm.entry_name,
                self.record_item_orm.specification_name,
                self.record_item_orm.record_id,
            ],
            select_stmt,
        )

        session.execute(stmt)

    def copy_from(
        self,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
        specification_names: Optional[Iterable[str]] = None,
        copy_entries: bool = False,
        copy_specifications: bool = False,
        copy_records: bool = False,
        *,
        session: Optional[Session] = None,
    ):
        """
        Copies entries, specifications, and record items from one dataset to another

        Entries and specifications are copied. Actual records are not duplicated - the new
        dataset will still refer to the same records as the source.

        Parameters
        ----------
        source_dataset_id
            ID of the dataset to copy from
        destination_dataset_id
            ID of the dataset to copy to
        entry_names
            Only copy records for these entries. If none, copy records for all entries
        specification_names
            Only copy records for these specifications. If none, copy records for all specifications
        copy_entries
            If True, copy entries from the source dataset.
        copy_entries
            If True, copy entries from the source dataset.
        copy_specifications
            If True, copy specifications from the source dataset.
        copy_records
            If True, copy record (items) from the source dataset. Implies copy_entries and copy_specifications.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:

            source_dataset_type = self.root_socket.datasets.lookup_type(source_dataset_id, session=session)

            if source_dataset_type != self.dataset_type:
                raise UserReportableError(
                    f"Source dataset type {source_dataset_type} does not match destination type {self.dataset_type}"
                )

            # Copy specifications
            if copy_specifications or copy_records:
                self.copy_specifications(
                    session, source_dataset_id, destination_dataset_id, specification_names=specification_names
                )

            # Copy entries
            if copy_entries or copy_records:
                self.copy_entries(session, source_dataset_id, destination_dataset_id, entry_names=entry_names)

            # Copy record items
            if copy_records:
                self.copy_record_items(
                    session,
                    source_dataset_id,
                    destination_dataset_id,
                    entry_names=entry_names,
                    specification_names=specification_names,
                )

    def clone(
        self,
        source_dataset_id: int,
        new_dataset_name: str,
        *,
        session: Optional[Session] = None,
    ):
        """
        Clones a dataset, copying all entries, specifications and record items

        Entries and specifications are copied. Actual records are not duplicated - the new
        dataset will still refer to the same records as the source.

        Parameters
        ----------
        source_dataset_id
            ID of the dataset to clone
        new_dataset_name
            Name of the new, cloned dataset
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            stmt = select(self.dataset_orm).where(BaseDatasetORM.id == source_dataset_id)
            stmt = stmt.options(noload("*"))
            source_orm = session.execute(stmt).scalar_one_or_none()

            if source_orm is None:
                raise MissingDataError(f"Cannot find dataset with ID {source_dataset_id} for cloning")

            ds_socket = self.root_socket.datasets.get_socket(source_orm.dataset_type)
            new_dataset_id = ds_socket.add(
                name=new_dataset_name,
                description=source_orm.description,
                tagline=source_orm.tagline,
                tags=source_orm.tags,
                provenance=source_orm.provenance,
                default_compute_tag=source_orm.default_compute_tag,
                default_compute_priority=source_orm.default_compute_priority,
                extras=source_orm.extras,
                creator_user=source_orm.creator_user,
                existing_ok=False,
            )

            #################################################
            # Copy entries, specifications, and record items
            #################################################
            self.copy_from(source_dataset_id, new_dataset_id, copy_records=True, session=session)

            #####################
            # Contributed values
            #####################
            # I changed my mind - maybe we shouldn't copy contributed values. But leaving this here
            # in case we want to in the future
            # stmt = text("""
            #    INSERT INTO contributed_values (dataset_id, name, values, index, values_structure, theory_level,
            #                                    units, theory_level_details, citations, external_url, doi, comments)
            #    SELECT :new_dataset_id, name, values, index, values_structure, theory_level,
            #           units, theory_level_details, citations, external_url, doi, comments
            #    FROM contributed_values
            #    WHERE dataset_id = :source_dataset_id
            # """)

            # stmt = stmt.bindparams(new_dataset_id=new_dataset_id, source_dataset_id=source_dataset_id)
            # session.execute(stmt)

            #####################
            # Attachments
            #####################
            # Similarly, I don't think we want to copy attachments. The whole point is that the new
            # dataset will be modified/expanded, so not sure it makes sense. But someone might want it
            # in the future

            return new_dataset_id



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from qcelemental.models import FailedOperation, ComputeError

from qcfractal.components.gridoptimization.testing_helpers import (
    submit_procedure_data as submit_go_procedure_data,
    generate_task_key as generate_go_task_key,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.torsiondrive.testing_helpers import (
    submit_procedure_data as submit_td_procdure_data,
    generate_task_key as generate_td_task_key,
)
from qcfractal.testing_helpers import run_service
from qcportal.record_models import RecordStatusEnum, PriorityEnum

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcfractal.db_socket import SQLAlchemySocket

# The slow ones are a bit more thorough - ie, some iterations are mixed complete/incomplete
test_files = [
    pytest.param("go_C4H4N2OS_mopac_pm6", marks=pytest.mark.slow),
    "td_H2O2_mopac_pm6",
    "go_H3NS_psi4_pbe",  # preopt = True
    "go_H2O2_psi4_pbe",  # preopt = False
]


def _submit_procedure_data(storage_socket, name: str, tag="*", priority=PriorityEnum.normal):
    if name.startswith("go"):
        return submit_go_procedure_data(storage_socket, name, tag, priority)
    else:
        return submit_td_procdure_data(storage_socket, name, tag, priority)


def _get_task_key_generator(name: str):
    if name.startswith("go"):
        return generate_go_task_key
    else:
        return generate_td_task_key


@pytest.mark.parametrize("procedure_file", test_files)
def test_record_client_reset_error_service(snowflake: QCATestingSnowflake, procedure_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    # create an alternative result dict where everything has errored
    failed_op = FailedOperation(
        error=ComputeError(error_type="test_error", error_message="this is just a test error"),
    )

    failed_data = {x: failed_op for x in result_data.keys()}

    run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

    # Make initial check for waiting service work
    storage_socket.records.reset_running([svc_id])

    while True:
        snowflake_client.reset_records([svc_id])

        with storage_socket.session_scope() as session:
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status == RecordStatusEnum.waiting

            # We should have also reset all the dependencies
            statuses = [x.record.status for x in rec.service.dependencies]
            assert all(x in [RecordStatusEnum.waiting, RecordStatusEnum.complete] for x in statuses)

            # Move the service to running. This will also use result_data to populate some of the
            # previously-errored (now waiting) dependencies
            run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

            session.expire(rec)
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status == RecordStatusEnum.running

            # Inject an error
            # needs multiple iterations - first generates tasks and submits results
            # second returns all tasks as errored
            # iteration will only happen when all tasks are completed or errored
            run_service(storage_socket, activated_manager_name, svc_id, keygen, failed_data, 200)

            session.expire(rec)
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status in [RecordStatusEnum.error, RecordStatusEnum.complete]

            # Should have errored dependencies
            if rec.status != RecordStatusEnum.complete:
                statuses = [x.record.status for x in rec.service.dependencies]
                assert RecordStatusEnum.error in statuses
            else:
                break


@pytest.mark.parametrize("procedure_file", test_files)
def test_record_client_cancel_waiting_service(snowflake: QCATestingSnowflake, procedure_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file, "test_tag", PriorityEnum.low)
    keygen = _get_task_key_generator(procedure_file)

    snowflake_client.cancel_records([svc_id])

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.cancelled
        assert rec.service is not None

        # Uncancel - will run?
        snowflake_client.uncancel_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.waiting
        assert rec.service is not None
        assert rec.service.compute_tag == "test_tag"
        assert rec.service.compute_priority == PriorityEnum.low

        finished, n_optimizations = run_service(
            storage_socket, activated_manager_name, svc_id, keygen, result_data, 200
        )

        assert finished

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete
        assert len(rec.compute_history) == 1
        assert len(rec.compute_history[0].outputs) == 1


@pytest.mark.parametrize("procedure_file", test_files)
def test_record_client_cancel_waiting_service_child(snowflake: QCATestingSnowflake, procedure_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

    # Cancel a child
    with storage_socket.session_scope() as s:
        ch_ids = storage_socket.records.get_children_ids(s, [svc_id])

    snowflake_client.cancel_records(ch_ids[0])

    # service is cancelled
    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.cancelled

        # Uncancel & continue
        snowflake_client.uncancel_records([svc_id])

        finished, n_optimizations = run_service(
            storage_socket, activated_manager_name, svc_id, keygen, result_data, 200
        )
        assert finished

        session.expire(rec)
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete


@pytest.mark.parametrize("procedure_file", test_files)
def test_record_client_cancel_running_service(snowflake: QCATestingSnowflake, procedure_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    # Get it running
    finished, n_optimizations = run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.running

        while not finished:
            meta = snowflake_client.cancel_records([svc_id])
            assert meta.n_updated == 1

            session.expire(rec)
            rec = session.get(BaseRecordORM, svc_id)

            assert rec.status == RecordStatusEnum.cancelled
            assert rec.service is not None  # service queue data left in place
            statuses = [x.record.status for x in rec.service.dependencies]
            assert all(x in [RecordStatusEnum.complete, RecordStatusEnum.cancelled] for x in statuses)
            changed_count = statuses.count(RecordStatusEnum.cancelled)
            assert meta.n_children_updated == changed_count

            # will it run after uncancel?
            meta = snowflake_client.uncancel_records([svc_id])
            assert meta.n_updated == 1
            assert meta.n_children_updated == changed_count

            session.expire(rec)
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status == RecordStatusEnum.waiting

            statuses = [x.record.status for x in rec.service.dependencies]
            assert all(x in [RecordStatusEnum.complete, RecordStatusEnum.waiting] for x in statuses)

            # we need two iterations. The first will move the service to running,
            # the second will actually iterate if necessary
            finished, n_optimizations = run_service(
                storage_socket, activated_manager_name, svc_id, keygen, result_data, 2
            )


@pytest.mark.parametrize("procedure_file", test_files)
def test_record_client_cancel_error_service(snowflake: QCATestingSnowflake, procedure_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    # create an alternative result dict where everything has errored
    failed_op = FailedOperation(
        error=ComputeError(error_type="test_error", error_message="this is just a test error"),
    )

    failed_data = {x: failed_op for x in result_data.keys()}

    run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

    with storage_socket.session_scope() as session:
        while True:
            snowflake_client.cancel_records([svc_id])

            session.expire_all()
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status == RecordStatusEnum.cancelled
            assert rec.service is not None  # service queue data left in place
            statuses = [x.record.status for x in rec.service.dependencies]
            assert all(
                x in [RecordStatusEnum.complete, RecordStatusEnum.error, RecordStatusEnum.cancelled] for x in statuses
            )

            # will it run after uncanceling and resetting?
            snowflake_client.uncancel_records([svc_id])
            snowflake_client.reset_records([svc_id])

            session.expire_all()
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status == RecordStatusEnum.waiting

            # We should have also reset all the dependencies
            statuses = [x.record.status for x in rec.service.dependencies]
            assert all(x in [RecordStatusEnum.waiting, RecordStatusEnum.complete] for x in statuses)

            # Move the service to running. This will also use result_data to populate some of the
            # previously-errored (now waiting) dependencies
            run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

            session.expire_all()
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status == RecordStatusEnum.running

            # Inject an error
            # needs multiple iterations - first generates tasks and submits results
            # second returns all tasks as errored
            # iteration will only happen when all tasks are completed or errored
            run_service(storage_socket, activated_manager_name, svc_id, keygen, failed_data, 200)

            session.expire_all()
            rec = session.get(BaseRecordORM, svc_id)
            assert rec.status in [RecordStatusEnum.error, RecordStatusEnum.complete]

            # Should have errored dependencies
            if rec.status != RecordStatusEnum.complete:
                statuses = [x.record.status for x in rec.service.dependencies]
                assert RecordStatusEnum.error in statuses
            else:
                break


@pytest.mark.parametrize("procedure_file", test_files)
def test_record_client_invalidate_completed_service(snowflake: QCATestingSnowflake, procedure_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    # Run it straight
    finished, n_optimizations = run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 200)

    assert finished

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete

        # Mark as invalid
        snowflake_client.invalidate_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.invalid

        # Uninvalidate
        snowflake_client.uninvalidate_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete

        # Invalidate a child
        with storage_socket.session_scope() as s:
            ch_ids = storage_socket.records.get_children_ids(s, [svc_id])

        snowflake_client.invalidate_records([ch_ids[0], ch_ids[1]])

        session.expire_all()
        rec = [session.get(BaseRecordORM, i) for i in [svc_id, ch_ids[0], ch_ids[1]]]
        assert rec[0].status == RecordStatusEnum.invalid
        assert rec[1].status == RecordStatusEnum.invalid
        assert rec[2].status == RecordStatusEnum.invalid

        # Uninvalidate one child - shouldn't uninvalidate service
        snowflake_client.uninvalidate_records([ch_ids[0]])

        session.expire_all()
        rec = [session.get(BaseRecordORM, i) for i in [svc_id, ch_ids[0], ch_ids[1]]]
        assert rec[0].status == RecordStatusEnum.invalid
        assert rec[1].status == RecordStatusEnum.complete
        assert rec[2].status == RecordStatusEnum.invalid

        # Uninvalidate service
        snowflake_client.uninvalidate_records([svc_id])

        session.expire_all()
        rec = [session.get(BaseRecordORM, i) for i in [svc_id, ch_ids[0], ch_ids[1]]]
        assert rec[0].status == RecordStatusEnum.complete
        assert rec[1].status == RecordStatusEnum.complete
        assert rec[2].status == RecordStatusEnum.complete


def get_children_ids(storage_socket: SQLAlchemySocket, svc_ids):
    with storage_socket.session_scope() as session:
        return storage_socket.records.get_children_ids(session, svc_ids)


@pytest.mark.parametrize("procedure_file", test_files)
@pytest.mark.parametrize("delete_children", [True, False])
def test_record_client_softdelete_service(snowflake: QCATestingSnowflake, procedure_file: str, delete_children: bool):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    with storage_socket.session_scope() as session:

        def check_children_deleted():
            ch_ids = get_children_ids(storage_socket, [svc_id])

            ch = [session.get(BaseRecordORM, i) for i in ch_ids]
            if delete_children:
                assert all(x.status == RecordStatusEnum.deleted for x in ch)
            else:
                assert all(x.status != RecordStatusEnum.deleted for x in ch)

        def check_children_undeleted():
            ch_ids = get_children_ids(storage_socket, [svc_id])

            ch = [session.get(BaseRecordORM, i) for i in ch_ids]
            assert all(x.status != RecordStatusEnum.deleted for x in ch)

        # 1. Service is waiting
        snowflake_client.delete_records([svc_id], soft_delete=True, delete_children=delete_children)

        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted

        snowflake_client.undelete_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.waiting

        # 2. running
        run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.running

        snowflake_client.delete_records([svc_id], soft_delete=True, delete_children=delete_children)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted
        assert rec.service is not None

        check_children_deleted()

        snowflake_client.undelete_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.waiting  # gets undeleted to "waiting"

        check_children_undeleted()

        # 3. error
        failed_op = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )
        failed_data = {x: failed_op for x in result_data.keys()}
        run_service(storage_socket, activated_manager_name, svc_id, keygen, failed_data, 3)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.error

        snowflake_client.delete_records([svc_id], soft_delete=True, delete_children=delete_children)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted
        assert rec.service is not None

        check_children_deleted()

        snowflake_client.undelete_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.error

        check_children_undeleted()

        # 4. cancelled
        snowflake_client.cancel_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.cancelled

        snowflake_client.delete_records([svc_id], soft_delete=True, delete_children=delete_children)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted
        assert rec.service is not None

        check_children_deleted()

        snowflake_client.undelete_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.cancelled

        check_children_undeleted()

        # 5. completed
        # reset and finish
        snowflake_client.uncancel_records([svc_id])
        snowflake_client.reset_records([svc_id])  # was error
        finished, n_optimizations = run_service(
            storage_socket, activated_manager_name, svc_id, keygen, result_data, 200
        )
        assert finished

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete
        assert rec.service is None

        # Now delete
        snowflake_client.delete_records([svc_id], soft_delete=True, delete_children=delete_children)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted

        check_children_deleted()

        snowflake_client.undelete_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete

        check_children_undeleted()

        # 6. Invalid
        snowflake_client.invalidate_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.invalid

        snowflake_client.delete_records([svc_id], soft_delete=True, delete_children=delete_children)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted

        check_children_deleted()

        snowflake_client.undelete_records([svc_id])

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.invalid

        check_children_undeleted()


@pytest.mark.parametrize("procedure_file", test_files)
@pytest.mark.parametrize("delete_children", [True, False])
def test_record_client_softdelete_service_child(
    snowflake: QCATestingSnowflake, procedure_file: str, delete_children: bool
):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 3)

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.running

        ch_ids = get_children_ids(storage_socket, [svc_id])
        snowflake_client.delete_records(ch_ids[0], soft_delete=True, delete_children=delete_children)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.deleted

        ch = [session.get(BaseRecordORM, i) for i in ch_ids]

        # which children were deleted
        if delete_children:
            assert all(x.status == RecordStatusEnum.deleted for x in ch)
        else:
            for c in ch:
                if c.id == ch_ids[0]:
                    assert c.status == RecordStatusEnum.deleted
                else:
                    assert c.status != RecordStatusEnum.deleted


@pytest.mark.parametrize("procedure_file", test_files)
@pytest.mark.parametrize("status", [RecordStatusEnum.waiting, RecordStatusEnum.running, RecordStatusEnum.complete])
@pytest.mark.parametrize("delete_children", [True, False])
def test_record_client_harddelete_service(
    snowflake: QCATestingSnowflake, procedure_file: str, status: RecordStatusEnum, delete_children: bool
):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    svc_id, result_data = _submit_procedure_data(storage_socket, procedure_file)
    keygen = _get_task_key_generator(procedure_file)

    with storage_socket.session_scope() as session:
        if status == RecordStatusEnum.waiting:
            ch_ids = get_children_ids(storage_socket, [svc_id])
            snowflake_client.delete_records([svc_id], soft_delete=False, delete_children=delete_children)

            rec = session.get(BaseRecordORM, svc_id)
            assert rec is None

            ch = [session.get(BaseRecordORM, i) for i in ch_ids]
            if delete_children:
                assert all(x is None for x in ch)
            else:
                assert all(x is not None for x in ch)
            return

        run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 1)

        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.running

        if status == RecordStatusEnum.running:
            ch_ids = get_children_ids(storage_socket, [svc_id])
            snowflake_client.delete_records(svc_id, soft_delete=False, delete_children=delete_children)

            session.expire_all()
            rec = session.get(BaseRecordORM, svc_id)
            assert rec is None

            ch = [session.get(BaseRecordORM, i) for i in ch_ids]
            if delete_children:
                assert all(x is None for x in ch)
            else:
                assert all(x is not None for x in ch)

            return

        run_service(storage_socket, activated_manager_name, svc_id, keygen, result_data, 200)

        session.expire_all()
        rec = session.get(BaseRecordORM, svc_id)
        assert rec.status == RecordStatusEnum.complete

        if status == RecordStatusEnum.complete:
            ch_ids = get_children_ids(storage_socket, [svc_id])
            snowflake_client.delete_records([svc_id], soft_delete=False, delete_children=delete_children)

            session.expire_all()
            rec = session.get(BaseRecordORM, svc_id)
            assert rec is None

            ch = [session.get(BaseRecordORM, i) for i in ch_ids]
            if delete_children:
                assert all(x is None for x in ch)
            else:
                assert all(x is not None for x in ch)

            return

        raise RuntimeError("Unhandled status in test")



"""
A command line interface to the qcfractal server.
"""

from __future__ import annotations

import argparse
import atexit
import logging
import multiprocessing
import os
import shutil
import signal
import sys
import textwrap
import threading
import time
import traceback
from typing import TYPE_CHECKING

import tabulate
import yaml

from qcfractal import __version__
from qcportal.auth import UserInfo
from .config import read_configuration, write_initial_configuration, FractalConfig, WebAPIConfig
from .db_socket.socket import SQLAlchemySocket
from .postgres_harness import PostgresHarness
from .process_targets import api_process, job_runner_process

if TYPE_CHECKING:
    from logging import Logger


class EndProcess(RuntimeError):
    pass


def pretty_bytes(num):
    # SO: https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%4.2f%sB" % (num, unit)
        num /= 1024.0
    return "%.2f%sB" % (num, "Yi")


def dump_config(qcf_config: FractalConfig, indent: int = 0) -> str:
    """
    Returns a string showing a full QCFractal configuration.

    It will be formatted in YAML. It will start and end with a line of '-'

    Parameters
    ----------
    qcf_config
        Configuration to print
    indent
        Indent the entire configuration by this many spaces

    Returns
    -------
    :
        The configuration as a human-readable string
    """

    s = "-" * 80 + "\n"
    cfg_str = yaml.dump(qcf_config.dict())
    s += textwrap.indent(cfg_str, " " * indent)
    s += "-" * 80
    return s


def start_database(config: FractalConfig, check_revision: bool) -> PostgresHarness:
    """
    Obtain a storage socket to a running postgres server

    If the server is not started and we are expected to manage it, this will also
    start it

    This returns a harness and a storage socket
    """

    logger = logging.getLogger(__name__)
    logger.info("Checking the PostgreSQL connection...")

    pg_harness = PostgresHarness(config.database)
    atexit.register(pg_harness.shutdown)

    if config.database.own and not pg_harness.postgres_initialized():
        raise RuntimeError("PostgreSQL instance has not been initialized?")

    # If we are expected to manage the postgres instance ourselves, start it
    # If not, make sure it is started
    pg_harness.ensure_alive()

    # Checks that the database exists
    if not pg_harness.can_connect():
        raise RuntimeError(f"Database at {config.database.safe_uri} does not exist?")

    # Check that the database is up to date
    if check_revision:
        SQLAlchemySocket.check_db_revision(config.database)

    return pg_harness


def parse_args() -> argparse.Namespace:
    """
    Sets up the command line arguments and parses them

    Returns
    -------
    :
        Argparse namespace containing the information about all the options specified on
        the command line
    """

    # Common help strings
    config_file_help = "Path to a QCFractal configuration file"
    verbose_help = "Output more details about the startup of qcfractal-server commands. Use twice for debug output"

    parser = argparse.ArgumentParser(description="A CLI for managing & running a QCFractal server.")

    parser.add_argument("--version", action="version", version=f"{__version__}")
    parser.add_argument("-v", "--verbose", action="count", default=0, help=verbose_help)

    parser.add_argument("--config", help=config_file_help)

    # Common arguments. These are added to the subcommands
    # They are similar to the global options above, but with SUPPRESS as the default
    base_parser = argparse.ArgumentParser(add_help=False)
    base_parser.add_argument("-v", "--verbose", action="count", default=argparse.SUPPRESS, help=verbose_help)
    base_parser.add_argument("--config", default=argparse.SUPPRESS, help=config_file_help)

    # Now start the real subcommands
    subparsers = parser.add_subparsers(dest="command")

    #####################################
    # init-config subcommand
    #####################################
    init_config = subparsers.add_parser(
        "init-config",
        help="Creates an initial configuration for a server",
        parents=[base_parser],
    )

    init_config.add_argument("--full", action="store_true", help="Create an example config will all fields")

    #####################################
    # init-db subcommand
    #####################################
    subparsers.add_parser(
        "init-db",
        help="Initializes a QCFractal server and database information from a given configuration",
        parents=[base_parser],
    )

    #####################################
    # start subcommand
    #####################################
    start = subparsers.add_parser("start", help="Starts a QCFractal server instance.", parents=[base_parser])

    # Allow some config settings to be altered via the command line
    start.add_argument("--port", type=int, help=WebAPIConfig.__pydantic_fields__["port"].description)
    start.add_argument("--host", type=str, help=WebAPIConfig.__pydantic_fields__["host"].description)
    start.add_argument("--logfile", type=str, help=FractalConfig.__pydantic_fields__["logfile"].description)
    start.add_argument("--loglevel", type=str, help=FractalConfig.__pydantic_fields__["loglevel"].description)
    start.add_argument(
        "--enable-security", type=bool, help=FractalConfig.__pydantic_fields__["enable_security"].description
    )

    start.add_argument(
        "--disable-job-runner",
        action="store_true",
        help="[ADVANCED] Disable the internal job runner (service updates and manager cleanup)",
    )

    #####################################
    # start-job-runner subcommand
    #####################################
    start_per = subparsers.add_parser(
        "start-job-runner", help="Starts a QCFractal server job-runner", parents=[base_parser]
    )

    # Allow some config settings to be altered via the command line
    start_per.add_argument("--logfile", type=str, help=FractalConfig.__pydantic_fields__["logfile"].description)
    start_per.add_argument("--loglevel", type=str, help=FractalConfig.__pydantic_fields__["loglevel"].description)

    #####################################
    # start-api subcommand
    #####################################
    start_api = subparsers.add_parser("start-api", help="Starts a QCFractal server instance.", parents=[base_parser])

    # Allow some config settings to be altered via the command line
    start_api.add_argument("--logfile", type=str, help=FractalConfig.__pydantic_fields__["logfile"].description)
    start_api.add_argument("--loglevel", type=str, help=FractalConfig.__pydantic_fields__["loglevel"].description)

    #####################################
    # upgrade-db subcommand
    #####################################
    subparsers.add_parser("upgrade-db", help="Upgrade QCFractal database.", parents=[base_parser])

    #####################################
    # upgrade-config subcommand
    #####################################
    subparsers.add_parser("upgrade-config", help="Upgrade a QCFractal configuration file.", parents=[base_parser])

    #####################################
    # info subcommand
    #####################################
    info = subparsers.add_parser(
        "info", help="Manage users and permissions on a QCFractal server instance.", parents=[base_parser]
    )
    info.add_argument(
        "category", nargs="?", default="server", choices=["server", "alembic"], help="The type of info to show"
    )

    #####################################
    # user subcommand
    #####################################
    user = subparsers.add_parser("user", help="Manage users for this instance", parents=[base_parser])

    # user sub-subcommands
    user_subparsers = user.add_subparsers(dest="user_command")

    # user list
    user_subparsers.add_parser("list", help="List information about all users", parents=[base_parser])

    # user info
    user_info = user_subparsers.add_parser("info", help="Show information about a user", parents=[base_parser])
    user_info.add_argument("username", default=None, type=str, help="The username to display information about.")

    # user add
    user_add = user_subparsers.add_parser("add", help="Add a user to the QCFractal server.", parents=[base_parser])
    user_add.add_argument("username", default=None, type=str, help="The username to add.")
    user_add.add_argument(
        "--password",
        type=str,
        required=False,
        help="The password for the user. If not specified, a default one will be created and printed.",
    )
    user_add.add_argument("--role", type=str, required=True, help="The role of this user on the server")

    user_add.add_argument(
        "--fullname", default=None, type=str, help="The real name or description of this user (optional)"
    )
    user_add.add_argument("--organization", default=None, type=str, help="The organization of this user (optional)")
    user_add.add_argument("--email", default=None, type=str, help="Email of the user (optional)")

    # user modify
    user_modify = user_subparsers.add_parser(
        "modify", help="Change a user's password or permissions.", parents=[base_parser]
    )
    user_modify.add_argument("username", default=None, type=str, help="The username to modify.")

    user_modify.add_argument(
        "--fullname", default=None, type=str, help="The real name or description of this user (optional)"
    )
    user_modify.add_argument("--organization", default=None, type=str, help="New organization of the user")
    user_modify.add_argument("--email", default=None, type=str, help="New email of the user")
    user_modify.add_argument("--role", default=None, type=str, help="New role of the user")

    user_modify_enable = user_modify.add_mutually_exclusive_group()
    user_modify_enable.add_argument("--enable", action="store_true", help="Enable this user")
    user_modify_enable.add_argument("--disable", action="store_true", help="Disable this user")

    user_modify_password = user_modify.add_mutually_exclusive_group()
    user_modify_password.add_argument(
        "--password", type=str, default=None, required=False, help="Change the user's password to the specified value."
    )
    user_modify_password.add_argument(
        "--reset-password",
        action="store_true",
        help="Reset the user's password. A new password will be generated and printed.",
    )

    # user delete
    user_delete = user_subparsers.add_parser("delete", help="Delete a user.", parents=[base_parser])
    user_delete.add_argument("username", default=None, type=str, help="The username to delete/remove")
    user_delete.add_argument("--no-prompt", action="store_true", help="Do not prompt for confirmation")

    #####################################
    # backup subcommand
    #####################################
    backup = subparsers.add_parser(
        "backup", help="Creates a postgres backup file of the current database.", parents=[base_parser]
    )
    backup.add_argument(
        "filename",
        type=str,
        help="The filename to dump the backup to",
    )

    #####################################
    # restore subcommand
    #####################################
    restore = subparsers.add_parser("restore", help="Restores the database from a backup file.", parents=[base_parser])
    restore.add_argument("filename", default=None, type=str, help="The filename to restore from.")

    args = parser.parse_args()
    return args


def server_init_config(config_path: str, full_config: bool):
    write_initial_configuration(config_path, full_config)

    print("*** Creating initial QCFractal configuration ***")
    print(f"Configuration path: {config_path}")
    print("NOTE: secret keys have been randomly generated. You likely won't need to change those")
    print(" !!! You will likely need/want to change some settings before")
    print("     initializing the database or starting the server !!!")


def server_init_db(config: FractalConfig):
    logger = logging.getLogger(__name__)
    logger.info("*** Initializing QCFractal database from configuration ***")

    pg_harness = PostgresHarness(config.database)
    atexit.register(pg_harness.shutdown)

    # If we own the database, initialize and start it
    if config.database.own and pg_harness.postgres_initialized():
        raise RuntimeError("PostgreSQL has been initialized or already exists")

    if config.database.own:
        pg_harness.initialize_postgres()

    pg_harness.create_database(create_tables=True)

    logger.info("QCFractal PostgreSQL instance is initialized")


def server_info(category: str, config: FractalConfig) -> None:
    # Just use raw printing here, rather than going through logging
    if category == "server":
        pg_harness = PostgresHarness(config.database)
        atexit.register(pg_harness.shutdown)
        pg_harness.ensure_alive()

        print()
        print("-" * 80)
        print("Python executable: ", sys.executable)
        print("QCFractal version: ", __version__)
        print("QCFractal alembic revision: ", pg_harness.get_alembic_version())

        if config.database.own:
            print("pg_ctl path: ", pg_harness._get_tool("pg_ctl"))

        print("PostgreSQL server version: ", pg_harness.get_postgres_version())
        print("-" * 80)
        print()
        print()
        print("Displaying QCFractal configuration below")
        print(dump_config(config))
    elif category == "alembic":
        print(f"Displaying QCFractal Alembic CLI configuration:\n")
        alembic_commands = SQLAlchemySocket.alembic_commands(config.database)
        print(" ".join(alembic_commands))


def setup_logging(config, logger: Logger):
    logger.info(f"QCFractal server base folder: {config.base_folder}")

    # Initialize the global logging infrastructure
    stdout_logging = True
    if config.logfile is not None:
        stdout_logging = False
        logger.info(f"Logging to {config.logfile} at level {config.loglevel}")
        log_handler = logging.FileHandler(config.logfile)
    else:
        logger.info(f"Logging to stdout at level {config.loglevel}")
        log_handler = logging.StreamHandler(sys.stdout)

    formatter = logging.Formatter(
        "[%(asctime)s] (%(processName)-16s) %(levelname)8s: %(name)s: %(message)s", "%Y-%m-%d %H:%M:%S %Z"
    )
    log_handler.setFormatter(formatter)

    # Reset the logger given the full configuration
    logging.getLogger().handlers = [log_handler]
    logging.getLogger().setLevel(config.loglevel)

    return stdout_logging


def server_start(config):
    logger = logging.getLogger(__name__)
    logger.info("*** Starting a QCFractal server ***")

    # Set up a multiprocessing context
    mp_context = multiprocessing.get_context("spawn")

    stdout_logging = setup_logging(config, logger)

    # Logger for the rest of this function
    logger = logging.getLogger(__name__)

    # Ensure that the database is alive, optionally starting it
    start_database(config, check_revision=True)

    # Set up a queue for logging. All child process will send logs
    # to this queue, and a separate thread will handle them
    logging_queue = mp_context.Queue()

    def _log_thread(queue):
        while True:
            record = queue.get()

            if record is None:
                # None is used as a sentinel to stop the thread
                break

            logging.getLogger(record.name).handle(record)

    log_thread = threading.Thread(target=_log_thread, args=(logging_queue,))
    log_thread.start()

    # Start up the api and job runner in separate processes
    api_proc = mp_context.Process(name="API_Process", target=api_process, args=(config, logging_queue, None))
    api_proc.start()

    job_runner_procs = [
        mp_context.Process(name=f"Job_Runner_{i}", target=job_runner_process, args=(config, logging_queue, None))
        for i in range(config.internal_job_processes)
    ]
    for p in job_runner_procs:
        p.start()

    def _cleanup(sig, frame):
        signame = signal.Signals(sig).name
        logger.debug("In cleanup of qcfractal_server")
        raise EndProcess(signame)

    signal.signal(signal.SIGINT, _cleanup)
    signal.signal(signal.SIGTERM, _cleanup)

    exitcode = 0
    try:
        while True:
            time.sleep(15)
            if not api_proc.is_alive():
                raise RuntimeError("API process died! Check the logs")
            if not all([p.is_alive() for p in job_runner_procs]):
                raise RuntimeError("A Job runner died! Check the logs")

    except EndProcess as e:
        if not stdout_logging:
            # Start logging to the screen again
            stdout_handler = logging.StreamHandler(sys.stdout)

            # Keep the old formatting (for consistency)
            stdout_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
            logging.getLogger().handlers.append(stdout_handler)

        logger.info("QCFractal server received EndProcess: " + str(e))
        logger.info("...stopping server...")

    except Exception as e:
        if not stdout_logging:
            # Start logging to the screen again
            stdout_handler = logging.StreamHandler(sys.stdout)

            # Keep the old formatting (for consistency)
            stdout_handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
            logging.getLogger().handlers.append(stdout_handler)

        tb = "".join(traceback.format_exception(None, e, e.__traceback__))
        logger.critical(f"Exception while running QCFractal server:\n{tb}")
        exitcode = 1

    api_proc.terminate()
    api_proc.join()

    for p in job_runner_procs:
        p.terminate()
        p.join()

    # Stop the logging thread
    logger.debug("Stopping logging thread")
    logging_queue.put(None)
    log_thread.join()

    sys.exit(exitcode)


def server_start_job_runner(config):
    from qcfractal.job_runner import FractalJobRunner

    logger = logging.getLogger(__name__)
    logger.info("*** Starting a QCFractal server job runner ***")

    setup_logging(config, logger)

    # Logger for the rest of this function
    # logger = logging.getLogger(__name__)

    # Ensure that the database is alive. This also handles checking stuff,
    # even if we don't own the db (which we shouldn't)
    start_database(config, check_revision=True)

    # Now just run the job runner directly
    job_runner = FractalJobRunner(config)
    job_runner.start()

    def _cleanup(sig, frame):
        logger.debug("In cleanup of job runner")
        job_runner.stop()

    signal.signal(signal.SIGINT, _cleanup)
    signal.signal(signal.SIGTERM, _cleanup)

    exit(0)


def server_start_api(config):
    from qcfractal.flask_app.waitress_app import FractalWaitressApp

    logger = logging.getLogger(__name__)
    logger.info("*** Starting a QCFractal API server ***")

    setup_logging(config, logger)

    # Logger for the rest of this function
    # logger = logging.getLogger(__name__)

    # Ensure that the database is alive. This also handles checking stuff,
    # even if we don't own the db (which we shouldn't)
    start_database(config, check_revision=True)

    # Now just run the api process
    api = FractalWaitressApp(config)
    api.run()


def server_upgrade_db(config):

    # Always set logging level to INFO, otherwise things are a bit quiet
    root_logger = logging.getLogger()
    if root_logger.level > logging.INFO:
        root_logger.setLevel(logging.INFO)

    logger = logging.getLogger(__name__)

    # Don't use start_database - we don't want to create the socket (which
    # will probably fail, because the version check should fail)
    pg_harness = PostgresHarness(config.database)
    atexit.register(pg_harness.shutdown)

    if config.database.own and not pg_harness.postgres_initialized():
        raise RuntimeError("PostgreSQL instance has not been initialized?")

    # If we are expected to manage the postgres instance ourselves, start it
    # If not, make sure it is started
    pg_harness.ensure_alive()

    # Checks that the database exists
    if not pg_harness.can_connect():
        raise RuntimeError(f"Database at {config.database.safe_uri} does not exist for upgrading?")

    logger.info(f"Upgrading the postgres database at {config.database.safe_uri}")

    try:
        SQLAlchemySocket.upgrade_database(config.database)
    except ValueError as e:
        print(str(e))
        sys.exit(1)


def server_upgrade_config(config_path):
    import secrets
    from qcfractal.old_config import OldFractalQCFConfig
    from qcfractal.config import convert_old_configuration

    logger = logging.getLogger(__name__)

    logger.info(f"Reading configuration data from {config_path}")
    with open(config_path, "r") as yf:
        file_data = yaml.safe_load(yf)

    # Is this really an old config?
    if "fractal" not in file_data:
        logger.info(f"Configuration appears to be up-to-date")
        return

    old_qcf_config = OldFractalQCFConfig(**file_data)
    new_qcf_config = convert_old_configuration(old_qcf_config)

    # Move the old file out of the way
    base_backup_path = config_path + ".backup"
    backup_path = base_backup_path

    # Find a suitable backup name. If the path exists,
    # add .1, .2, .3, etc, until it does
    i = 1
    while os.path.exists(backup_path):
        backup_path = base_backup_path + f".{i}"
        i += 1

    logger.info(f"Moving original configuration to {backup_path}")
    shutil.move(config_path, backup_path)

    # We strip out some stuff from the configuration, including things that were
    # set to the default
    new_qcf_config_dict = new_qcf_config.dict(skip_defaults=True)
    new_qcf_config_dict["database"].pop("base_folder")
    new_qcf_config_dict.pop("base_folder")

    # Generate secret keys for the user. They can always change it later
    new_qcf_config_dict["api"]["secret_key"] = secrets.token_hex(24)
    new_qcf_config_dict["api"]["jwt_secret_key"] = secrets.token_hex(24)

    logger.info(f"Writing new configuration data to  {config_path}")
    with open(config_path, "w") as yf:
        yaml.dump(new_qcf_config_dict, yf)

    logger.info("*" * 80)
    logger.info("Your configuration file has been upgraded, but will most likely need some fine tuning.")
    logger.info(f"Please edit {config_path} by hand. See the upgrade documentation for details.")
    logger.info("*" * 80)


def server_user(args: argparse.Namespace, config: FractalConfig):
    user_command = args.user_command

    # Don't check revision here - it will be done in the SQLAlchemySocket constructor
    start_database(config, check_revision=False)
    storage = SQLAlchemySocket(config)

    def print_user_info(u: UserInfo):
        enabled = "True" if u.enabled else "False"
        print("-" * 80)
        print(f"      username: {u.username}")
        print(f"     full name: {u.fullname}")
        print(f"  organization: {u.organization}")
        print(f"         email: {u.email}")
        print(f"          role: {u.role}")
        print(f"       enabled: {enabled}")
        print(f"     auth type: {u.auth_type}")
        print("-" * 80)

    if user_command == "list":
        user_list = storage.users.list()

        table_rows = []

        for u in user_list:
            u_obj = UserInfo(**u)
            table_rows.append((u_obj.username, u_obj.role, u_obj.auth_type, u_obj.enabled, u_obj.fullname))

        print()
        table_str = tabulate.tabulate(table_rows, headers=["username", "role", "auth type", "enabled", "fullname"])
        print(table_str)
        print()

    if user_command == "info":
        u = storage.users.get(args.username)
        print_user_info(UserInfo(**u))

    if user_command == "add":
        fullname = args.fullname if args.fullname is not None else ""
        organization = args.organization if args.organization is not None else ""
        email = args.email if args.email is not None else ""

        user_info = UserInfo(
            username=args.username,
            role=args.role,
            fullname=fullname,
            email=email,
            organization=organization,
            enabled=True,
        )

        pw = storage.users.add(user_info, args.password)

        if args.password is None:
            print(f"\nAutogenerated password for {args.username} is below")
            print("-" * 80)
            print(pw)
            print("-" * 80)

        print(f"Created user {args.username}")

    if user_command == "modify":
        u = storage.users.get(args.username)

        if args.fullname is not None:
            u["fullname"] = args.fullname
        if args.organization is not None:
            u["organization"] = args.organization
        if args.email is not None:
            u["email"] = args.email
        if args.role is not None:
            u["role"] = args.role

        # Enable/disable are separate on the command line
        if args.enable is True:
            u["enabled"] = True
        if args.disable is True:
            u["enabled"] = False

        print(f"Updating information for user {u['username']}")
        storage.users.modify(UserInfo(**u), as_admin=True)

        # Passwords are handled separately
        if args.reset_password is True:
            print(f"Resetting password...")
            pw = storage.users.change_password(u["username"], None)
            print(f"\nNew autogenerated password for {args.username} is below")
            print("-" * 80)
            print(pw)
            print("-" * 80)
        elif args.password is not None:
            print("Setting the password...")
            storage.users.change_password(u["username"], args.password)

    if user_command == "delete":
        u = storage.users.get(args.username)
        print_user_info(UserInfo(**u))

        if args.no_prompt is not True:
            r = input("Really delete this user? (Y/N): ")
            if r.lower() == "y":
                storage.users.delete(args.username)
                print("Deleted!")
            else:
                print("ABORTING!")
        else:
            storage.users.delete(args.username)
            print("Deleted!")


def server_backup(args: argparse.Namespace, config: FractalConfig):
    pg_harness = start_database(config, check_revision=True)

    db_size = pg_harness.database_size()
    pretty_size = pretty_bytes(db_size)

    print("\n")
    print(f"Backing up the database at {config.database.safe_uri}")
    print(f"Current database size is {pretty_size}")
    print("\n")

    # Bigger than 250GB?
    if db_size > 250 * 2**30:
        print("\n" + "*" * 80)
        print("This is a pretty big database! This will take a while...")
        print("Consider alternate backup strategies, such as pg_basebackup and WAL archiving")
        print("*" * 80 + "\n")

    filepath = os.path.realpath(args.filename)
    filepath_tmp = filepath + ".in_progress"

    if os.path.exists(filepath):
        raise RuntimeError(f"File {filepath} already exists. Will not overwrite!")
    if os.path.exists(filepath_tmp):
        raise RuntimeError(
            f"File {filepath_tmp} already exists. This is a leftover temporary file, and probably "
            f"needs to be deleted, but it could also mean you have a backup running already"
        )

    print(f"Backing up to {filepath_tmp}...", end=None)

    def _remove_temporary_file():
        if os.path.exists(filepath_tmp):
            print("NOTE: Removing temporary file due to unexpected exit")
            os.remove(filepath_tmp)

    atexit.register(_remove_temporary_file)

    pg_harness.backup_database(filepath_tmp)

    print("Done!")
    print(f"Moving to {filepath}")

    # Check again...
    if os.path.exists(filepath):
        raise RuntimeError(
            f"File {filepath} already exists. It didn't before! {filepath_tmp} contains the backup we "
            f"just made. You are on your own."
        )

    shutil.move(filepath_tmp, filepath)

    print("Backup complete!")
    atexit.unregister(_remove_temporary_file)


def server_restore(args: argparse.Namespace, config: FractalConfig):
    if not os.path.isfile(args.filename):
        raise RuntimeError(f"Backup file {args.filename} does not exist or is not a file!")

    logger = logging.getLogger(__name__)
    logger.info("Checking the PostgreSQL connection...")

    pg_harness = PostgresHarness(config.database)
    atexit.register(pg_harness.shutdown)

    # If we own the db, it might not have been initialized
    # (ie, postgres server hasn't been set up at all)
    if config.database.own and not pg_harness.postgres_initialized():
        pg_harness.initialize_postgres()

    pg_harness.ensure_alive()

    # If we are expected to manage the postgres instance ourselves, start it
    # If not, make sure it is started
    pg_harness.ensure_alive()

    db_exists = pg_harness.can_connect()

    if db_exists:
        print("!WARNING! This will erase old data!")
        print("Run 'qcfractal-server backup' before this option if you want to keep it,")
        print("or copy the database to somewhere else")

        user_required_input = f"REMOVEALLDATA {str(config.database.database_name)}"
        print(f"!WARNING! If you are sure you wish to proceed please type '{user_required_input}' below.")
        inp = input("> ")
        print()

        if inp != user_required_input:
            print(f"Input does not match. Aborting")
            return

        pg_harness.delete_database()

    restore_size = os.path.getsize(args.filename)
    pretty_size = pretty_bytes(restore_size)

    print(f"\nSize of the backup file: {pretty_size}")
    print("Starting restore...", end=None)

    pg_harness.restore_database(args.filename)

    print("done")
    print("\nRestore complete!")


def main():
    # Parse all the command line arguments
    args = parse_args()

    # Set up a log handler. This is used before the logfile is set up
    log_handler = logging.StreamHandler(sys.stdout)

    # If the user wants verbose output (particularly about startup of all the commands), then set logging level
    # to be DEBUG
    # Use a stripped down format, since we are just logging to stdout
    if args.verbose == 0:
        logging.basicConfig(level=logging.WARNING, handlers=[log_handler], format="%(levelname)s: %(message)s")
    elif args.verbose == 1:
        logging.basicConfig(level=logging.INFO, handlers=[log_handler], format="%(levelname)s: %(name)s: %(message)s")
    elif args.verbose >= 2:
        logging.basicConfig(level=logging.DEBUG, handlers=[log_handler], format="%(levelname)s: %(name)s: %(message)s")

    logger = logging.getLogger(__name__)

    # If command is not 'init', we need to build the configuration
    # Priority: 1.) command line
    #           2.) environment variables
    #           3.) config file
    # We do not explicitly handle environment variables.
    # They are handled automatically by pydantic

    # Handle some arguments given on the command line
    # They are only valid if starting a server
    cmd_config = {"api": {}}

    # command = the subcommand used (init, start, etc)
    if args.command == "start":
        if args.port is not None:
            cmd_config["api"]["port"] = args.port
        if args.host is not None:
            cmd_config["api"]["host"] = args.host
        if args.logfile is not None:
            cmd_config["logfile"] = args.logfile
        if args.loglevel is not None:
            cmd_config["loglevel"] = args.loglevel
        if args.enable_security is not None:
            cmd_config["enable_security"] = args.enable_security

    if args.command in ["start-job-runner", "start-api"]:
        if args.logfile is not None:
            cmd_config["logfile"] = args.logfile
        if args.loglevel is not None:
            cmd_config["loglevel"] = args.loglevel

    ###############################################################
    # Shortcuts here for initializing/upgrading the configuration
    # We don't want to read old configs with new code, or the
    #    config doesn't exist yet
    ###############################################################

    if args.command == "init-config":
        if args.config is None:
            raise RuntimeError("Configuration file path (--config) is required for initialization")
        server_init_config(args.config, args.full)
        exit(0)

    if args.command == "upgrade-config":
        if args.config is None:
            raise RuntimeError("Configuration file path (--config) is required for upgrading configuration")
        server_upgrade_config(args.config)
        exit(0)

    # Check for the config path on the command line. The command line
    # always overrides environment variables
    config_paths = []
    if args.config is not None:
        config_paths.append(args.config)
    elif "QCF_CONFIG_PATH" in os.environ:
        config_paths.append(os.getenv("QCF_CONFIG_PATH"))

    config_paths = [os.path.abspath(x) for x in config_paths]

    # Now read and form the complete configuration
    qcf_config = read_configuration(config_paths, cmd_config)

    cfg_str = dump_config(qcf_config, 4)
    logger.debug("Assembled the following configuration:\n" + cfg_str)

    if args.command == "info":
        server_info(args.category, qcf_config)
    elif args.command == "init-db":
        server_init_db(qcf_config)
    elif args.command == "start":
        server_start(qcf_config)
    elif args.command == "start-job-runner":
        server_start_job_runner(qcf_config)
    elif args.command == "start-api":
        server_start_api(qcf_config)
    elif args.command == "upgrade-db":
        server_upgrade_db(qcf_config)
    elif args.command == "user":
        server_user(args, qcf_config)
    elif args.command == "backup":
        server_backup(args, qcf_config)
    elif args.command == "restore":
        server_restore(args, qcf_config)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.gridoptimization.record_db_models import GridoptimizationRecordORM
from qcfractal.components.optimization.record_db_models import OptimizationRecordORM
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.testing_helpers import run_service
from qcportal.auth import UserInfo
from qcportal.gridoptimization import (
    GridoptimizationSpecification,
    GridoptimizationKeywords,
    compare_gridoptimization_records,
)
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification, SinglepointProtocols
from qcportal.utils import now_at_utc
from .testing_helpers import (
    compare_gridoptimization_specs,
    test_specs,
    load_procedure_data,
    generate_task_key,
    load_record_data,
)

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal.managers import ManagerName
    from sqlalchemy.orm.session import Session


@pytest.mark.parametrize("spec", test_specs)
def test_gridoptimization_socket_add_get(
    storage_socket: SQLAlchemySocket, session: Session, spec: GridoptimizationSpecification
):
    hooh = load_molecule_data("peroxide2")
    h3ns = load_molecule_data("go_H3NS")

    time_0 = now_at_utc()
    meta, ids = storage_socket.records.gridoptimization.add([hooh, h3ns], spec, "tag1", PriorityEnum.low, None, True)
    time_1 = now_at_utc()
    assert meta.success

    recs = [session.get(GridoptimizationRecordORM, i) for i in ids]

    assert len(recs) == 2
    for r in recs:
        assert r.record_type == "gridoptimization"
        assert r.status == RecordStatusEnum.waiting
        assert compare_gridoptimization_specs(spec, r.specification.model_dict())

        # Service queue entry should exist with the proper tag and priority
        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert recs[0].initial_molecule.identifiers["molecule_hash"] == hooh.get_hash()
    assert recs[1].initial_molecule.identifiers["molecule_hash"] == h3ns.get_hash()


def test_gridoptimization_socket_find_existing_1(storage_socket: SQLAlchemySocket):
    spec = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            scans=[
                {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"},
                {"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-90, 0], "step_type": "absolute"},
            ],
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        ),
    )

    hooh = load_molecule_data("peroxide2")
    meta, id1 = storage_socket.records.gridoptimization.add([hooh], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.gridoptimization.add([hooh], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_gridoptimization_socket_find_existing_2(storage_socket: SQLAlchemySocket):
    # some modifications to the input specification
    spec1 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            preoptimization=True,
            scans=[
                {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"},
                {"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-90, 0], "step_type": "absolute"},
            ],
        ),
        optimization_specification=OptimizationSpecification(
            program="optprog1",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(),
            qc_specification=QCSpecification(
                program="prog2",
                driver="deferred",
                method="b3lyp",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(wavefunction="none"),
            ),
        ),
    )

    spec2 = GridoptimizationSpecification(
        program="gridoptimization",
        keywords=GridoptimizationKeywords(
            scans=[
                {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"},
                {"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-90, 0], "step_type": "absolute"},
            ],
        ),
        optimization_specification=OptimizationSpecification(
            program="optPROG1",
            keywords={"k": "value"},
            qc_specification=QCSpecification(
                program="prOG2",
                driver="deferred",
                method="b3LYP",
                basis="6-31g",
                keywords={"k2": "values2"},
                protocols=SinglepointProtocols(stdout=True),
            ),
        ),
    )

    mol1 = load_molecule_data("go_H3NS")
    mol2 = load_molecule_data("peroxide2")
    meta, id1 = storage_socket.records.gridoptimization.add([mol1, mol2], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 2
    assert meta.inserted_idx == [0, 1]

    meta, id2 = storage_socket.records.gridoptimization.add([mol1, mol2], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 2
    assert meta.existing_idx == [0, 1]
    assert id1 == id2


@pytest.mark.parametrize(
    "test_data_name",
    [
        "go_C4H4N2OS_mopac_pm6",
        "go_H2O2_psi4_b3lyp",
        "go_H3NS_psi4_pbe",
    ],
)
def test_gridoptimization_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName, test_data_name: str
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data(test_data_name)

    storage_socket.users.add(UserInfo(username="submit_user", role="submit", enabled=True))

    meta_1, id_1 = storage_socket.records.gridoptimization.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, "submit_user", True
    )
    id_1 = id_1[0]
    assert meta_1.success

    time_0 = now_at_utc()
    finished, n_optimizations = run_service(
        storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 100
    )
    time_1 = now_at_utc()

    assert finished is True

    rec = session.get(GridoptimizationRecordORM, id_1)

    assert rec.status == RecordStatusEnum.complete
    assert time_0 < rec.modified_on < time_1
    assert len(rec.compute_history) == 1
    assert len(rec.compute_history[-1].outputs) == 1
    assert rec.compute_history[-1].status == RecordStatusEnum.complete
    assert time_0 < rec.compute_history[-1].modified_on < time_1
    assert rec.service is None

    desc_info = storage_socket.records.get_short_descriptions([id_1])[0]
    short_desc = desc_info["description"]
    assert desc_info["record_type"] == rec.record_type
    assert desc_info["created_on"] == rec.created_on
    assert rec.specification.program in short_desc
    assert rec.specification.optimization_specification.program in short_desc
    assert rec.specification.optimization_specification.qc_specification.program in short_desc
    assert rec.specification.optimization_specification.qc_specification.method in short_desc

    out = rec.compute_history[-1].outputs["stdout"].get_output()
    assert "Grid optimization finished successfully!" in out

    assert len(rec.optimizations) == n_optimizations

    for o in rec.optimizations:
        optr = session.get(OptimizationRecordORM, o.optimization_id)
        assert optr.energies[-1] == o.energy


def test_gridoptimization_socket_run_duplicate(
    storage_socket: SQLAlchemySocket,
    session: Session,
    activated_manager_name: ManagerName,
):
    input_spec_1, molecules_1, result_data_1 = load_procedure_data("go_H2O2_psi4_b3lyp")

    meta_1, id_1 = storage_socket.records.gridoptimization.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, True
    )
    id_1 = id_1[0]
    assert meta_1.success

    run_service(storage_socket, activated_manager_name, id_1, generate_task_key, result_data_1, 100)

    rec_1 = session.get(GridoptimizationRecordORM, id_1)
    assert rec_1.status == RecordStatusEnum.complete
    opt_ids_1 = [x.optimization_id for x in rec_1.optimizations]

    # Submit again, without duplicate checking
    meta_2, id_2 = storage_socket.records.gridoptimization.add(
        [molecules_1], input_spec_1, "test_tag", PriorityEnum.low, None, False
    )
    id_2 = id_2[0]
    assert meta_2.success
    assert id_2 != id_1

    run_service(storage_socket, activated_manager_name, id_2, generate_task_key, result_data_1, 1000)

    rec_2 = session.get(GridoptimizationRecordORM, id_2)
    assert rec_2.status == RecordStatusEnum.complete
    opt_ids_2 = [x.optimization_id for x in rec_2.optimizations]

    assert set(opt_ids_1).isdisjoint(opt_ids_2)


def test_gridoptimization_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "go_C4H4N2OS_mopac_pm6",
        "go_H2O2_psi4_pbe",
        "go_H2O2_psi4_b3lyp",
        "go_H3NS_psi4_pbe",
        "go_error_17761737",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_gridoptimizations(ins_ids[0], include=["**"])
        rec_1.fetch_children(include=["**"], force_fetch=True)
        compare_gridoptimization_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"
        for opt in rec_1.optimizations.values():
            assert opt.creator_user == "submit_user"



"""Remove server stats log

Revision ID: 73b4838a6839
Revises: 75b80763e901
Create Date: 2024-05-06 10:54:44.383709

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "73b4838a6839"
down_revision = "75b80763e901"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_server_stats_log_timestamp", table_name="server_stats_log", postgresql_using="brin")
    op.drop_table("server_stats_log")

    op.execute("DELETE FROM internal_jobs WHERE name = 'update_server_stats'")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "server_stats_log",
        sa.Column("id", sa.INTEGER(), autoincrement=True, nullable=False),
        sa.Column("timestamp", postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=False),
        sa.Column("collection_count", sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column("molecule_count", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("record_count", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("outputstore_count", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("access_count", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("db_total_size", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("db_table_size", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("db_index_size", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("db_table_information", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
        sa.Column("error_count", sa.BIGINT(), autoincrement=False, nullable=True),
        sa.Column("service_queue_status", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
        sa.Column("task_queue_status", postgresql.JSON(astext_type=sa.Text()), autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint("id", name="server_stats_log_pkey"),
    )
    op.create_index(
        "ix_server_stats_log_timestamp", "server_stats_log", ["timestamp"], unique=False, postgresql_using="brin"
    )
    # ### end Alembic commands ###



from __future__ import annotations

import logging
from typing import Dict, Any, Tuple, Callable

from qcelemental.models import Molecule
from sqlalchemy import select

from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket import SQLAlchemySocket
from qcfractalcompute.compress import compress_result
from qcportal.compression import decompress, CompressionEnum
from qcportal.managers import ManagerName
from qcportal.record_models import RecordStatusEnum, RecordTask

mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="2234-5678-1234-5678")


class DummyJobProgress:
    """
    Functor for updating progress and cancelling internal jobs

    This is a dummy version used for testing
    """

    def __init__(self):
        self._runner_uuid = "1234-5678-9101-1213"

    def update_progress(self, progress: int):
        pass

    def cancelled(self) -> bool:
        return False

    def deleted(self) -> bool:
        return False


def run_service(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    record_id: int,
    task_key_generator: Callable,
    result_data: Dict[str, Any],
    max_iterations: int = 20,
) -> Tuple[bool, int]:
    """
    Runs a service
    """

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, record_id)
        assert rec.status in [RecordStatusEnum.waiting, RecordStatusEnum.running]

        creator_user = rec.creator_user.username if rec.creator_user is not None else None

        tag = rec.service.compute_tag
        priority = rec.service.compute_priority
        service_id = rec.service.id

    n_records = 0
    n_iterations = 0
    finished = False

    while n_iterations < max_iterations:
        with storage_socket.session_scope() as session:
            n_services = storage_socket.services.iterate_services(session)

            # iterate_services will handle errors
            if n_services == 0:
                finished = True
                break

            # Kinda hacky...
            # Run any internal jobs that iterate_services added
            jobname = f"iterate_service_{service_id}"
            stmt = select(InternalJobORM).where(InternalJobORM.unique_name == jobname)
            job_orm = session.execute(stmt).scalar_one_or_none()

            if job_orm is not None:
                storage_socket.internal_jobs._run_single(
                    session, job_orm, logging.getLogger("internal_job"), DummyJobProgress()
                )
                # The function that iterates a service returns True if it is finished
                if job_orm.result is True:
                    rec: BaseRecordORM = session.get(BaseRecordORM, record_id)

                    if rec.status == RecordStatusEnum.error:
                        print("Error in service dependency")
                        print(rec.status)
                        print(rec.compute_history[-1].status)
                        print(decompress(rec.compute_history[-1].outputs["error"].data, CompressionEnum.zstd))

                    assert rec.status == RecordStatusEnum.complete
                    assert rec.service is None
                    finished = True
                    break

        n_iterations += 1

        with storage_socket.session_scope() as session:
            rec = session.get(BaseRecordORM, record_id)
            assert rec.status == RecordStatusEnum.running

        # only do 5 tasks at a time. Tests iteration when stuff is not completed
        manager_programs = storage_socket.managers.get([manager_name.fullname])[0]["programs"]
        manager_tasks_d = storage_socket.tasks.claim_tasks(manager_name.fullname, manager_programs, ["*"], limit=5)
        manager_tasks = [RecordTask(**x) for x in manager_tasks_d]

        # Sometimes a task may be duplicated in the service dependencies.
        # The C8H6 test has this "feature"
        ids = set(x.record_id for x in manager_tasks)

        with storage_socket.session_scope() as session:
            recs = [session.get(BaseRecordORM, i) for i in ids]
            all_usernames = [x.creator_user.username if x.creator_user is not None else None for x in recs]
            assert all(x == creator_user for x in all_usernames)
            assert all(x.task.compute_priority == priority for x in recs)
            assert all(x.task.compute_tag == tag for x in recs)

        manager_ret = {}
        for t in manager_tasks:
            # The results dict has keys that are generated by a function
            # That same function is passed into this function
            task_key = task_key_generator(t)

            task_result = result_data.get(task_key, None)
            if task_result is None:
                raise RuntimeError(f"Cannot find task results! key = {task_key}")

            manager_ret[t.id] = compress_result(task_result.dict())

        rmeta = storage_socket.tasks.update_finished(manager_name.fullname, manager_ret)
        assert rmeta.n_accepted == len(manager_tasks)
        n_records += len(manager_ret)

    return finished, n_records


def compare_validate_molecule(m1: Molecule, m2: Molecule) -> bool:
    """
    Validates, and then compares molecules

    Molecules get validated when added to the server, so if we are comparing
    molecules, we often need to validate the input as well.
    """

    m1_v = Molecule(**m1.dict(), validate=True)
    m2_v = Molecule(**m2.dict(), validate=True)
    return m1_v == m2_v



from qcarchivetesting import load_hash_test_data
from qcfractal.components.reaction.record_db_models import ReactionSpecificationORM
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.reaction import ReactionSpecification, ReactionKeywords
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols

common_sp_spec = QCSpecification(
    program="prog2",
    driver=SinglepointDriver.energy,
    method="",
    basis="def2-tzvp",
    keywords={"k": "value"},
    protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
)

common_opt_spec = OptimizationSpecification(
    program="optprog2",
    keywords={"k": "value"},
    protocols=OptimizationProtocols(trajectory="none"),
    qc_specification=common_sp_spec.copy(),
)


def test_reaction_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("reaction_specification_tests")

    # Hashes are independent of opt and sp spec
    # TODO - protocols are not part of model
    for t in test_data.values():
        del t["protocols"]
    spec_map = [
        (
            k,
            ReactionSpecification(
                singlepoint_specification=common_sp_spec, optimization_specification=common_opt_spec, **v
            ),
        )
        for k, v in test_data.items()
    ]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.reaction.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(ReactionSpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_reaction_socket_add_specification_same_1(spec_test_runner):
    # both sp and opt
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    spec_test_runner("reaction", spec1, spec1, True)


def test_reaction_socket_add_specification_same_2(spec_test_runner):
    # only sp
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=None,
    )

    spec_test_runner("reaction", spec1, spec1, True)


def test_reaction_socket_add_specification_same_3(spec_test_runner):
    # only opt
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=None,
        optimization_specification=common_opt_spec,
    )

    spec_test_runner("reaction", spec1, spec1, True)


def test_reaction_socket_add_specification_diff_1(spec_test_runner):
    # removing opt spec
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    spec2 = spec1.copy(update={"optimization_specification": None})
    spec_test_runner("reaction", spec1, spec2, False)


def test_reaction_socket_add_specification_diff_2(spec_test_runner):
    # removing qc spec
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    spec2 = spec1.copy(update={"singlepoint_specification": None})
    spec_test_runner("reaction", spec1, spec2, False)


def test_reaction_socket_add_specification_diff_3(spec_test_runner):
    # removing qc spec
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=None,
        optimization_specification=common_opt_spec,
    )

    spec2 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=None,
    )

    spec_test_runner("reaction", spec1, spec2, False)


def test_reaction_socket_add_specification_diff_4(spec_test_runner):
    # different basis in qc spec
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    qc_spec = spec1.singlepoint_specification.copy(update={"basis": "sto-3g"})
    spec2 = spec1.copy(update={"singlepoint_specification": qc_spec})
    spec_test_runner("reaction", spec1, spec2, False)


def test_reaction_socket_add_specification_diff_5(spec_test_runner):
    # different basis in opt spec
    spec1 = ReactionSpecification(
        program="reaction",
        keywords=ReactionKeywords(),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    qc_spec = spec1.optimization_specification.qc_specification.copy(update={"basis": "sto-3g"})
    opt_spec = spec1.optimization_specification.copy(update={"qc_specification": qc_spec})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("reaction", spec1, spec2, False)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import select

from qcportal.auth import is_valid_groupname, GroupInfo
from qcportal.exceptions import UserManagementError
from .db_models import GroupORM

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, List, Dict, Any, Union


class GroupSocket:
    """
    Socket for managing groups
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

    def _get_internal(self, session: Session, groupname_or_id: Union[int, str]) -> GroupORM:
        """
        Obtain the ORM for a particular group.

        If the group is not found, an exception is raised. The ORM is attached to the given session

        Parameters
        ----------
        session
            SQLAlchemy session to use for querying

        Returns
        -------
        :
            ORM of the specified group
        """

        if isinstance(groupname_or_id, int) or groupname_or_id.isdecimal():
            stmt = select(GroupORM).where(GroupORM.id == groupname_or_id)
        else:
            is_valid_groupname(groupname_or_id)
            stmt = select(GroupORM).where(GroupORM.groupname == groupname_or_id)

        group = session.execute(stmt).scalar_one_or_none()

        if group is None:
            raise UserManagementError(f"Group {groupname_or_id} does not exist.")

        return group

    def list(self, *, session: Optional[Session] = None) -> List[Dict[str, Any]]:
        """
        Get information about all groups

        Parameters
        ----------
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session, True) as session:
            stmt = select(GroupORM).order_by(GroupORM.id.asc())
            all_groups = session.execute(stmt).scalars().all()
            return [x.model_dict() for x in all_groups]

    def get(self, groupname_or_id: Union[int, str], *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Obtains information for a group

        Parameters
        ----------
        groupname_or_id
            The name or ID of the group
        session
            An existing SQLAlchemy session to use. If None, one will be created
        """

        with self.root_socket.optional_session(session, True) as session:
            group = self._get_internal(session, groupname_or_id)
            return group.model_dict()

    def add(self, group_info: GroupInfo, *, session: Optional[Session] = None):
        """
        Adds a new group

        Parameters
        ----------
        group_info
            Information about the new group
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        # Should have been checked already, but defense in depth
        is_valid_groupname(group_info.groupname)

        # ID should not be set
        if group_info.id is not None:
            raise UserManagementError("Cannot add a group - id was given as part of new group info")

        try:
            with self.root_socket.optional_session(session) as session:
                group = GroupORM(groupname=group_info.groupname, description=group_info.description)
                session.add(group)
        except IntegrityError:
            raise UserManagementError(f"Group {group_info.groupname} already exists")

        self._logger.info(f"Group {group_info.groupname} added")

    def delete(self, groupname_or_id: Union[int, str], *, session: Optional[Session] = None) -> None:
        """Removes a group

        This will raise an exception if the group doesn't exist or is being referenced elsewhere in the
        database.

        Parameters
        ----------
        groupname_or_id
            The name or ID of the group
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        try:
            with self.root_socket.optional_session(session) as session:
                group = self._get_internal(session, groupname_or_id)
                session.delete(group)
        except IntegrityError:
            raise UserManagementError("Group could not be deleted. Likely it is being referenced somewhere")

        self._logger.info(f"Group {groupname_or_id} deleted")



from __future__ import annotations

from typing import TYPE_CHECKING

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_serverinfo_client_access_logged(secure_snowflake_allow_read: QCATestingSnowflake):
    time_0 = now_at_utc()
    client = secure_snowflake_allow_read.client("admin_user", test_users["admin_user"]["pw"])
    read_client = secure_snowflake_allow_read.client()

    client.query_access_log()
    client.query_molecules(molecular_formula=["C"])

    read_client.get_molecules([123], missing_ok=True)
    time_0 = now_at_utc()

    # This will return 6, because the requests to /login and /information was done in constructing the clients
    query_res = client.query_access_log()

    # creating the client can add pings
    accesses = list(query_res)
    assert len(accesses) >= 6

    assert accesses[5].module == "auth"
    assert accesses[5].full_uri == "/auth/v1/login"
    assert accesses[5].user == "admin_user"

    assert accesses[4].module == "api"
    assert accesses[4].full_uri == "/api/v1/information"
    assert accesses[4].user == "admin_user"

    assert accesses[3].module == "api"
    assert accesses[3].full_uri == "/api/v1/information"
    assert accesses[3].user is None

    assert accesses[2].module == "api"
    assert accesses[2].full_uri == "/api/v1/access_logs/query"
    assert accesses[2].user == "admin_user"

    assert accesses[1].module == "api"
    assert accesses[1].full_uri == "/api/v1/molecules/query"
    assert accesses[1].user == "admin_user"

    assert accesses[0].module == "api"
    assert accesses[0].full_uri == "/api/v1/molecules/bulkGet"
    assert accesses[0].user is None

    assert accesses[0].response_bytes > 0
    assert accesses[1].response_bytes > 0
    assert accesses[2].response_bytes > 0
    assert accesses[3].response_bytes > 0
    assert accesses[0].request_bytes > 0
    assert accesses[1].request_bytes > 0
    assert accesses[2].request_bytes > 0
    assert accesses[3].request_bytes == 0


def test_serverinfo_client_access_not_logged(postgres_server, pytestconfig):
    pg_harness = postgres_server.get_new_harness("serverinfo_client_access_not_logged")
    encoding = pytestconfig.getoption("--client-encoding")

    with QCATestingSnowflake(pg_harness, encoding, log_access=False) as server:
        client = server.client()
        client.query_access_log()
        client.query_molecules(molecular_formula=["C"])

        client.get_molecules([])

        # This will return 0 because logging is disabled
        query_res = client.query_access_log()
        assert len(list(query_res)) == 0


def test_serverinfo_client_access_delete(snowflake_client: PortalClient):
    time_0 = now_at_utc()
    snowflake_client.query_access_log()
    time_12 = now_at_utc()
    snowflake_client.query_molecules(molecular_formula=["C"])
    time_23 = now_at_utc()
    snowflake_client.get_molecules([123], missing_ok=True)
    time_4 = now_at_utc()

    query_res = snowflake_client.query_access_log(after=time_0)
    query_res_l = list(query_res)
    assert len(query_res_l) == 3

    # Delete anything related to constructing the client
    snowflake_client.delete_access_log(time_0)

    n_deleted = snowflake_client.delete_access_log(time_12)
    assert n_deleted == 1

    n_deleted = snowflake_client.delete_access_log(time_12)
    assert n_deleted == 0

    n_deleted = snowflake_client.delete_access_log(time_4)
    assert n_deleted == 2

    # All of the above generated accesses!
    n_deleted = snowflake_client.delete_access_log(now_at_utc())
    assert n_deleted == 7



"""change manager programs column type

Revision ID: 93fa1a1b70f8
Revises: 5ae71b845526
Create Date: 2025-04-27 10:28:22.184783

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "93fa1a1b70f8"
down_revision = "5ae71b845526"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "compute_manager",
        "programs",
        existing_type=postgresql.JSON(astext_type=sa.Text()),
        type_=postgresql.JSONB(astext_type=sa.Text()),
        existing_nullable=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "compute_manager",
        "programs",
        existing_type=postgresql.JSONB(astext_type=sa.Text()),
        type_=postgresql.JSON(astext_type=sa.Text()),
        existing_nullable=False,
    )
    # ### end Alembic commands ###
